"""Just in time compiled tools (seperated from tools so we don't keep
compiling them"""
import numpy as np
from numba import jit, boolean


@jit(cache=True, nopython=True, nogil=True)
def mask_ring_median(values_array, positions_array, alpha):  # pragma: no cover
    """Find outlier pixels in a single ring via a single pass with the median.

    Parameters
    ----------
    values_array : ndarray
        The ring values
    positions_array : ndarray
        The positions of the values
    alpha: float
        The threshold

    Returns
    -------
    removals: np.ndarray
        The positions of pixels to be removed from the data
    """
    z = np.abs(values_array - np.median(values_array)) / np.std(values_array)
    removals = positions_array[np.where(z > alpha)]
    return removals


@jit(cache=True, nopython=True, nogil=True)
def mask_ring_mean(values_array, positions_array, alpha):  # pragma: no cover
    """Find outlier pixels in a single ring via a pixel by pixel method with
    the mean.

    Parameters
    ----------
    values_array : ndarray
        The ring values
    positions_array : ndarray
        The positions of the values
    alpha: float
        The threshold

    Returns
    -------
    removals: np.ndarray
        The positions of pixels to be removed from the data
    """
    m = np.ones(positions_array.shape, dtype=boolean)
    removals = []
    while True:
        b = np.array(
            [item in removals for item in positions_array], dtype=boolean
        )
        m[b] = False
        v = values_array[m]
        if len(v) <= 1:
            break
        std = np.std(v)
        if std == 0.0:
            break
        norm_v_list = np.abs(v - np.mean(v)) / std
        if np.all(norm_v_list < alpha):
            break
        # get the index of the worst pixel
        worst_idx = np.argmax(norm_v_list)
        # add the worst position to the mask
        removals.append(positions_array[m][worst_idx])
    return removals



def load_data(filename, minrows=10, **kwargs):
    """Find and load data from a text file.

    The data reading starts at the first matrix block of at least minrows rows
    and constant number of columns.  This seems to work for most of the
    datafiles including those generated by PDFGetX2.

    filename -- name of the file we want to load data from.
    minrows  -- minimum number of rows in the first data block.
                All rows must have the same number of floating point values.
    usecols  -- zero-based index of columns to be loaded, by default use
                all detected columns.  The reading skips data blocks that
                do not have the usecols-specified columns.
    unpack   -- return data as a sequence of columns that allows tuple
                unpacking such as  x, y = loadData(FILENAME, unpack=True).
                Note transposing the loaded array as loadData(FILENAME).T
                has the same effect.
    kwargs   -- keyword arguments that are passed to numpy.loadtxt

    Return a numpy array of the data.
    See also numpy.loadtxt for more details.
    """
    from numpy import array, loadtxt
    # determine the arguments
    delimiter = kwargs.get('delimiter')
    usecols = kwargs.get('usecols')
    # required at least one column of floating point values
    mincv = (1, 1)
    # but if usecols is specified, require sufficient number of columns
    # where the used columns contain floats
    if usecols is not None:
        hiidx = max(-min(usecols), max(usecols) + 1)
        mincv = (hiidx, len(set(usecols)))

    # Check if a line consists of floats only and return their count
    # Return zero if some strings cannot be converted.
    def countcolumnsvalues(line_):
        try:
            words = line_.split(delimiter)
            # remove trailing blank columns
            while words and not words[-1].strip():
                words.pop(-1)
            nc = len(words)
            if usecols is not None:
                nv = len([float(words[i]) for i in usecols])
            else:
                nv = len([float(w) for w in words])
        except (IndexError, ValueError):
            nc = nv = 0
        return nc, nv

    # make sure fid gets cleaned up
    with open(filename, 'rb') as fid:
        # search for the start of datablock
        start = ncvblock = None
        fpos = (0, 0)
        nrows = 0
        for line in fid:
            fpos = (fpos[1], fpos[1] + len(line))
            line = line.decode()
            ncv = countcolumnsvalues(line)
            if ncv < mincv:
                start = None
                continue
            # ncv is acceptable here, require the same number of columns
            # throughout the datablock
            if start is None or ncv != ncvblock:
                ncvblock = ncv
                nrows = 0
                start = fpos[0]
            nrows += 1
            # block was found here!
            if nrows >= minrows:
                break
        # Return an empty array when no data found.
        # loadtxt would otherwise raise an exception on loading from EOF.
        if start is None:
            rv = array([], dtype=float)
        else:
            fid.seek(start)
            # always use usecols argument so that loadtxt does not crash
            # in case of trailing delimiters.
            kwargs.setdefault('usecols', list(range(ncvblock[0])))
            rv = loadtxt(fid, **kwargs)
    return rv



"""Visualize the data."""
from typing import Union, Any

import numpy as np
from matplotlib.axes import Axes
from numpy import ndarray

__all__ = [
    'auto_label',
    'auto_text',
    'normalize',
    'plot_line',
    'plot_fit',
    'shift',
    'text_position',
    'set_minor_tick'
]

_LABELS = {
    "xy": (r"Q ($\mathrm{\AA}^{-1}$)", r"I (A. U.)"),
    "chi": (r"Q ($\mathrm{\AA}^{-1}$)", r"I (A. U.)"),
    "iq": (r"Q ($\mathrm{\AA}^{-1}$)", r"I (A. U.)"),
    "sq": (r"Q ($\mathrm{\AA}^{-1}$)", r"S"),
    'fq': (r"Q ($\mathrm{\AA}^{-1}$)", r"F ($\mathrm{\AA}^{-1}$)"),
    "gr": (r"r ($\mathrm{\AA}$)", r"G ($\mathrm{\AA}^{-2}$)"),
    "fgr": (r"r ($\mathrm{\AA}$)", r"G ($\mathrm{\AA}^{-2}$)")
}
_TEXT_XY = (0.8, 0.8)


def normalize(vis_data: ndarray):
    """Normalize the data so that the max(data[n]) - min(data[n]) = 1 for n > 0.

    Parameters
    ----------
    vis_data : ndarray
        The data to visualize. The first row is independent variables and the later rows are dependent variables.
    """
    vis_data = np.copy(vis_data)
    vis_data[1:] /= np.max(vis_data[1]) - np.min(vis_data[1])
    return vis_data


def shift(vis_data: ndarray, ax: Axes, gap: float = 0, inplace: bool = False):
    """Shift the data so that the max(data[n-1]) - max(data[n]) = line_spacing for n > 1.

    Parameters
    ----------
    vis_data : ndarray
        The data to visualize. The first row is independent variables and the later rows are dependent variables.

    ax : Axes
        The axes to plot the data on.

    gap : float
        The spacing between two adjacent lines.

    inplace : bool
        If True, shift the data inplace. Else, shift the copied data.
    """
    if not inplace:
        vis_data = np.copy(vis_data)
    lines = ax.get_lines()
    if len(lines) > 0:
        last_y = lines[-1].get_ydata()
        y = vis_data[1:]
        vis_data[1:] += np.min(last_y) - np.max(y) - gap
    return vis_data


def plot_line(vis_data: ndarray, ax: Axes, xy_kwargs: dict = None):
    """Plot the first two rows of the data.

    Parameters
    ----------
    vis_data : ndarray
        The data to visualize. The first row is independent variables and the later rows are dependent variables.

    ax : Axes
        The axes to plot the data on.

    xy_kwargs : dict
        The kwargs for the ax.plot().
    """
    if xy_kwargs is None:
        xy_kwargs = {}
    ax.plot(vis_data[0], vis_data[1], **xy_kwargs)
    return ax


def text_position(vis_data: ndarray, text_xy: tuple) -> tuple:
    """Calculate the position of the text in data coordinates."""
    fx, fy = text_xy
    x1, y1 = vis_data[:2]
    xa = (1 - fx) * np.min(x1) + fx * np.max(x1)
    ya = (1 - fy) * np.min(y1) + fy * np.max(y1)
    return xa, ya


def complimentary(color: Union[str, Any]):
    """Return the complimentary color of a color string."""
    if color[0] != "#":
        from matplotlib.colors import to_hex
        hexcolor = to_hex(color)[1:]
    else:
        hexcolor = color[1:]
    # convert the string into hex
    intcolor = int(hexcolor, 16)
    # invert the three bytes
    # as good as substracting each of RGB component by 255(FF)
    comp_color = 0xFFFFFF ^ intcolor
    # convert the color back to hex by prefixing a #
    comp_color = "#{:06X}".format(comp_color)
    # return the result
    return comp_color


def auto_text(text: str, ax: Axes, text_xy: tuple = None, index: int = -1):
    """
    Automatically add the text.

    Parameters
    ----------
    text : str
        The content of the text.

    ax : Axes
        The axes to plot the data on.

    text_xy : tuple
        The relative position. The horizontal bound is the xlim. The vertical is the two curves.

    index : int
        The index of the lines to add text. Default -1 (the last line).
    """
    lines = ax.get_lines()
    if len(lines) == 0:
        return ax
    vis_data = lines[index].get_data()
    color = lines[index].get_color()
    if text_xy is None:
        text_xy = _TEXT_XY
    xy = text_position(vis_data, text_xy)
    ax.annotate(text, xy, ha="left", va="center", color=color)
    return ax


def auto_label(ax: Axes, label_type: str):
    """Automatically label the axes according to the name of the vis_field.

    Parameters
    ----------
    ax : Axes
        The axes to plot the data on.

    label_type : str
        The field in the input dictionary. Its value is the data for visualization.
    """
    if label_type in _LABELS:
        labels = _LABELS.get(label_type)
        ax.set_xlabel(labels[0])
        ax.set_ylabel(labels[1])
    return ax


def get_yzero(y: ndarray, ycalc: ndarray, ydiff: ndarray) -> ndarray:
    """Get the base line yzero of ydiff so that max(yzero + ydiff) = min(min(y), min(ydiff))"""
    return (min(float(np.min(y)), float(np.min(ycalc))) - float(np.max(ydiff))) * np.ones_like(ydiff)


def plot_fit(vis_data: ndarray, ax: Axes,
             xy_kwargs: dict = None, xycalc_kwargs: dict = None, xydiff_kwargs: dict = None,
             xyzero_kwargs: dict = None, fill_kwargs: dict = None,
             yzero: ndarray = None):
    """Visualize the fit.

    Parameters
    ----------
    vis_data : ndarray
        The data to visualize. The first three rows are independent variable, dependent variable, and fitting.

    ax : Axes
        The axes to plot on.

    xy_kwargs : dict
        The kwargs for plotting the y v.s. x curve.

    xycalc_kwargs : dict
        The kwargs for plotting the ycalc v.s. x curve.

    xydiff_kwargs : dict
        The kwargs for plotting the ydiff v.s. x curve.

    xyzero_kwargs : dict
        The kwargs for plotting the yzero v.s. x curve.

    fill_kwargs : dict
        The kwargs for filling in the area between ydiff and yzero.

    yzero : ndarray
        The base line corresponding to the zero value in ydiff.
    """
    # use the default value if None
    if xyzero_kwargs is None:
        xyzero_kwargs = {}
    if xydiff_kwargs is None:
        xydiff_kwargs = {}
    if xycalc_kwargs is None:
        xycalc_kwargs = {}
    if xy_kwargs is None:
        xy_kwargs = {}
    if fill_kwargs is None:
        fill_kwargs = {}
    # split data
    if len(vis_data.shape) != 2:
        raise ValueError('Invalid data shape: {}. Need 2D data array.'.format(vis_data.shape))
    if vis_data.shape[0] < 3:
        raise ValueError('Invalid data dimension: {}. Need dim >= 3'.format(vis_data.shape[0]))
    x, y, ycalc = vis_data[:3]
    ydiff = y - ycalc
    # shift ydiff
    if yzero is None:
        yzero = get_yzero(y, ycalc, ydiff)
    ydiff += yzero
    # circle data curve
    _xy_kwargs = {'fillstyle': 'none', 'label': 'data'}
    _xy_kwargs.update(xy_kwargs)
    data_line, = ax.plot(x, y, 'o', **_xy_kwargs)
    # solid calculation curve
    _xycalc_kwargs = {'label': 'fit', 'color': complimentary(data_line.get_color())}
    _xycalc_kwargs.update(xycalc_kwargs)
    ax.plot(x, ycalc, '-', **_xycalc_kwargs)
    # dash zero difference curve
    _xyzero_kwargs = {'color': 'grey'}
    _xyzero_kwargs.update(xyzero_kwargs)
    ax.plot(x, yzero, '--', **_xyzero_kwargs)
    # solid shifted difference curve
    _xydiff_kwargs = {'label': 'residuals', 'color': data_line.get_color()}
    _xydiff_kwargs.update(xydiff_kwargs)
    diff_line, = ax.plot(x, ydiff, '-', **_xydiff_kwargs)
    # fill in area between curves
    if fill_kwargs.pop('fill', True):
        _fill_kwargs = {'alpha': 0.4, 'color': diff_line.get_color()}
        _fill_kwargs.update(fill_kwargs)
        ax.fill_between(x, ydiff, yzero, **_fill_kwargs)
    return ax


def set_minor_tick(ax: Axes, n: int = 2):
    """Set one minor tick between major ticks."""
    from matplotlib.ticker import AutoMinorLocator
    ax.xaxis.set_minor_locator(AutoMinorLocator(n))
    ax.yaxis.set_minor_locator(AutoMinorLocator(n))
    return ax



import matplotlib.pyplot as plt
import numpy as np
import pytest

import pdfstream.visualization.tools as tools


def test_complimentary():
    assert tools.complimentary("black") == "#FFFFFF"


def test_auto_text():
    fig = plt.figure()
    ax = fig.add_subplot(111)
    tools.auto_text("nothing", ax)


@pytest.mark.parametrize(
    "wrong_data",
    [np.zeros((4, 5, 1)), np.zeros((2, 5))]
)
def test_plot_fit_eorr(wrong_data):
    fig = plt.figure()
    ax = fig.add_subplot(111)
    with pytest.raises(ValueError):
        tools.plot_fit(wrong_data, ax=ax)



import numpy as np
import pytest
from diffpy.pdfgetx import PDFConfig

import pdfstream.callbacks.from_start as mod
from pdfstream.data import ni_dspacing_file


def test_query_ai(db_with_dark_and_light):
    db = db_with_dark_and_light
    start = mod.get_start_of_run(db[-1])
    mod.query_ai(start, "calibration_md")


@pytest.mark.parametrize(
    "det_name, shape",
    [
        ("pe1_image", (2048, 2048)),
        pytest.param("pe1_stats1_total", None, marks=pytest.mark.xfail)
    ]
)
def test_get_img_from_run(db_with_dark_and_light, det_name, shape):
    db = db_with_dark_and_light
    img = mod.get_img_from_run(db[-1], det_name)
    assert img.shape == shape


@pytest.mark.parametrize(
    "dk_id_key, shape",
    [
        ("sc_dk_field_uid", (2048, 2048)),
        ("a_key_not_existing", None)
    ]
)
def test_query_dk_img(db_with_dark_and_light, dk_id_key, shape):
    db = db_with_dark_and_light
    start = mod.get_start_of_run(db[-1])
    dk_img = mod.query_dk_img(start, det_name="pe1_image", db=db, dk_id_key=dk_id_key)
    if shape:
        assert isinstance(dk_img, np.ndarray)
        assert dk_img.shape == shape
    else:
        assert dk_img is None


@pytest.mark.parametrize(
    "start, composition_key, wavelength_key, expect",
    [
        (
            {'sample_composition': {'Ni': 1.0}},
            'sample_composition',
            'bt_wavelength',
            ('Ni1.0', None)
        ),
        (
            {'bt_wavelength': 0.16},
            'sample_composition',
            'bt_wavelength',
            ('Ni', 0.16)
        ),
        (
            {'sample_composition': 'C60'},
            'sample_composition',
            'bt_wavelength',
            ('C60', None)
        )
    ]
)
def test_query_bt_info(start, composition_key, wavelength_key, expect):
    config = mod.query_bt_info(
        start,
        composition_key=composition_key,
        wavelength_key=wavelength_key,
        default_composition="Ni"
    )
    pdfconfig = PDFConfig(**config)
    assert (pdfconfig.composition, pdfconfig.wavelength) == expect


def test_query_bg_img(db_with_img_and_bg_img, test_data):
    db = db_with_img_and_bg_img
    run = db[-1]
    bg_img = mod.query_bg_img(
        run.metadata['start'],
        bkgd_sample_name_key="bkgd_sample_name",
        sample_name_key="sample_name",
        det_name="pe1_image",
        db=db,
        dk_id_key="sc_dk_field_uid"
    )
    assert np.array_equal(bg_img, np.ones_like(test_data["Ni_img"]))


@pytest.mark.parametrize(
    "start, expect",
    [
        (
            {"detector": "perkin_elmer", "bt_wavelength": 0.18, "sample_name": "Ni_calib"},
            {"detector": "perkin_elmer", "wavelength": '0.18', "calibrant": str(ni_dspacing_file)}
        )
    ]
)
def test_get_calib_info(start, expect):
    dct = mod.get_calib_info(start, "detector", "bt_wavelength", "sample_name")
    assert dct == expect



"""The input / output functions related to file system."""
from datetime import datetime
from pathlib import Path
from typing import Dict, Any

import fabio
import pyFAI
import yaml
from numpy import ndarray

from pdfstream.vend.loaddata import load_data


def load_ai_from_poni_file(poni_file: str) -> pyFAI.AzimuthalIntegrator:
    """Initiate the AzimuthalIntegrator using poni file."""
    ai = pyFAI.load(poni_file)
    return ai


def load_ai_from_calib_result(calib_result: dict) -> pyFAI.AzimuthalIntegrator:
    """Initiate the AzimuthalIntegrator using calibration information."""
    ai = pyFAI.azimuthalIntegrator.AzimuthalIntegrator()
    # different from poni file, set_config only accepts dictionary of lowercase keys
    _calib_result = _lower_key(calib_result)
    # the pyFAI only accept strings so the None should be parsed to a string
    _calib_result = _str_none(_calib_result)
    # the old version of poni uses dist intead of distance and the new version only recognize "distance"
    if ("dist" in _calib_result) and "distance" not in _calib_result:
        _calib_result["distance"] = _calib_result["dist"]
    ai.set_config(_calib_result)
    return ai


def load_img(img_file: str) -> ndarray:
    """Load the img data from the img_file."""
    img = fabio.open(img_file).data
    return img


def write_img(filepath: str, img: ndarray, template: str) -> None:
    """Write out the image data as the same type of the template file."""
    temp_img = fabio.open(template)
    temp_img.data = img
    temp_img.save(filepath)
    return


def load_array(data_file: str, minrows=10, **kwargs) -> ndarray:
    """Load data columns from the .txt file and turn columns to rows and return the numpy array."""
    return load_data(data_file, minrows=minrows, **kwargs).T


def load_dict_from_poni(poni_file: str) -> dict:
    """Turn the poni file to pyFAI readable dictionary."""
    with Path(poni_file).open('r') as f:
        geometry = yaml.safe_load(f)
    return _lower_key(geometry)


def _lower_key(dct: Dict[str, Any]) -> Dict[str, Any]:
    """Return dictionary with all keys in lower case."""
    return {key.lower(): value for key, value in dct.items()}


def _str_none(dct: Dict[str, Any]) -> Dict[str, Any]:
    """Make all the None value to string 'none'."""
    return {key: "none" if value is None else value for key, value in dct.items()}


def server_message(msg: str):
    """Print a message to a uniform format of server message."""
    t = datetime.now().strftime("%x %X")
    print("[{}] {}".format(t, msg))



from pathlib import Path
from tempfile import TemporaryDirectory

import matplotlib.pyplot as plt
import pytest

import pdfstream.transformation.cli as cli


@pytest.mark.parametrize(
    "parallel, plot_setting",
    [(False, None), (True, "OFF")]
)
def test_transform(test_data, parallel, plot_setting):
    with TemporaryDirectory() as temp_dir:
        dcts = cli.transform(
            test_data["Ni_gr_file"], test_data["Ni_chi_file"],
            plot_setting=plot_setting, output_dir=temp_dir, parallel=parallel, test=True
        )
        plt.close()
        for dct in dcts:
            for output_file in dct.values():
                assert Path(output_file).is_file()



"""The functions used in the command line interface. The input and output are all files."""
import typing as tp
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor, as_completed
from pathlib import Path, PurePath

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.axes import Axes

import pdfstream.integration as integ
import pdfstream.io as io
import pdfstream.visualization as vis


def integrate(
    poni_file: str,
    *img_files: str,
    bg_img_file: str = None,
    mask_file: str = None,
    output_dir: str = ".",
    bg_scale: float = None,
    mask_setting: tp.Union[dict, str] = None,
    integ_setting: dict = None,
    plot_setting: tp.Union[dict, str] = None,
    img_setting: tp.Union[dict, str] = None,
    parallel: bool = False,
    test: bool = False
) -> tp.List[str]:
    """Conduct azimuthal integration on the two dimensional diffraction images.

    The image will be first subtracted by background if background image file is given. Then, it will be binned
    in azimuthal direction according to the geometry provided by the poni file. The pixels far away from the
    average in each bin will be masked. The mask will be applied on the background subtracted image and the
    image will be integrated again by the pyFAI. The polarization correction and pixel-splitting algorithm will
    be applied according to user settings before the integration. The results are saved as chi files.

    Parameters
    ----------
    poni_file : str
        The path to the poni file. It will be read by pyFAI.

    img_files : str
        The arbitrary number of paths to the image file. It will be read by fabio.

    bg_img_file : str
        The path to the background image file. It should have the same dimension as the data image. If None,
        no background subtraction will be done.

    mask_file : str
        The .npy file of the mask array. Use pyFai convention. 1 are masked pixels, 0 are good pixels.

    output_dir : str
        The directory to save the chi data file. Default current working directory.

    bg_scale : float
        The scale of the background. Default 1

    mask_setting : dict or 'OFF'
        The settings for the auto-masking. See the arguments for mask_img (
        https://xpdacq.github.io/xpdtools/xpdtools.html?highlight=mask_img#xpdtools.tools.mask_img). To turn
        off the auto masking, enter "OFF".

    integ_setting : dict
        The settings for the integration. See the arguments for integrate1d (
        https://pyfai.readthedocs.io/en/latest/api/pyFAI.html#module-pyFAI.azimuthalIntegrator).

    plot_setting : dict or 'OFF'
        The keywords for the matplotlib.pyplot.plot (
        https://matplotlib.org/3.2.1/api/_as_gen/matplotlib.pyplot.plot.html). To turn off the plotting,
        enter "OFF".

    img_setting : dict or 'OFF'
        The keywords for the matplotlib.pyplot.imshow (
        https://matplotlib.org/3.2.1/api/_as_gen/matplotlib.pyplot.imshow.html). Besides, there is a key
        'z_score', which determines the range of the colormap. The range is mean +/- z_score * std in the
        statistics of the image. To turn of the image, enter "OFF".

    parallel : bool
        If True, run the processing in multiple process.

    test : bool
        If True, run in test mode (for developers).

    Returns
    -------
    chi_files : a list of strings
        The path to the output chi file.
    """
    Path(output_dir).mkdir(parents=True, exist_ok=True)
    if parallel:
        executor = ProcessPoolExecutor() if not test else ThreadPoolExecutor()
        jobs = [
            executor.submit(
                _integrate, poni_file, img_file, bg_img_file=bg_img_file, mask_file=mask_file,
                output_dir=output_dir, bg_scale=bg_scale, mask_setting=mask_setting, integ_setting=integ_setting,
                plot_setting=plot_setting, img_setting=img_setting, test=test
            )
            for img_file in img_files
        ]
        return [job.result() for job in as_completed(jobs)]
    return [
        _integrate(
            poni_file, img_file, bg_img_file=bg_img_file, mask_file=mask_file,
            output_dir=output_dir, bg_scale=bg_scale, mask_setting=mask_setting, integ_setting=integ_setting,
            plot_setting=plot_setting, img_setting=img_setting, test=test
        )
        for img_file in img_files
    ]


def _integrate(
    poni_file: str,
    img_file: str,
    bg_img_file: str = None,
    mask_file: str = None,
    output_dir: str = ".",
    bg_scale: float = None,
    mask_setting: tp.Union[dict, str] = None,
    integ_setting: dict = None, plot_setting: tp.Union[dict, str] = None,
    img_setting: tp.Union[dict, str] = None,
    test: bool = False
) -> str:
    """Sub-function for integrate."""
    if integ_setting is None:
        integ_setting = dict()
    ai = io.load_ai_from_poni_file(poni_file)
    bg_img = io.load_img(bg_img_file) if bg_img_file else None
    mask = np.load(mask_file) if mask_file else None
    img = io.load_img(img_file)
    chi_name = Path(img_file).with_suffix('.chi').name
    chi_path = Path(output_dir).joinpath(chi_name)
    integ_setting.update({'filename': str(chi_path)})
    integ.get_chi(
        ai, img, bg_img=bg_img, mask=mask, bg_scale=bg_scale, mask_setting=mask_setting,
        integ_setting=integ_setting, plot_setting=plot_setting, img_setting=img_setting,
    )
    if not test:
        plt.show()
    return chi_path


def average(out_file: str, *img_files, weights: tp.List[float] = None) -> None:
    """Average the single channel image files with weights.

    Parameters
    ----------
    out_file : str
        The output file path. It will be the type as the first image in img_files.

    img_files : a tuple of str
        The image files to be averaged.

    weights : an iterable of floats
        The weights for the images. If None, images will not be weighted when averaged.
    """
    img_files: tp.Tuple[str]
    imgs = (io.load_img(_) for _ in img_files)
    avg_img = integ.avg_imgs(imgs, weights=weights)
    # make the directory if not exists
    out_file_path = Path(out_file)
    if not out_file_path.parent.is_dir():
        out_file_path.parent.mkdir(parents=True)
    # write out the file
    io.write_img(out_file, avg_img, img_files[0])
    return


def waterfall(
    *data_files: str, ax: Axes = None, mode: str = "line", normal: bool = True,
    stack: bool = True, gap: float = 0, texts: tp.List[str] = None, text_xy: tuple = None,
    label: str = None, minor_tick: tp.Union[int, None] = 2, legends: tp.List[str] = None,
    colors: tp.Iterable = None, show_fig: bool = True, **kwargs
) -> Axes:
    """Visualize the data in multiple data files in a waterfall or comparison plot.

    The data must be multi-columns matrix in a txt file. A header can be included in the file and it won't be
    read. The plots will be stacked in a waterfall if stack = True or overlapping together if stack = False. The
    visualization has different modes for different kinds of plotting. Currently, it supports

    'line' mode
        Each dataset is a single curve of x and y data.

    'fit' mode
        Each dataset contains three curves, one curve for data, one curve for the fit and a difference curve
        below it.

    Parameters
    ----------
    data_files : a list of file paths
        The file paths. Each file contains column data. The required format depends on the 'mode'.
        If mode = 'line', data = (x_array, y_array)
        If mode = 'fit', data = (x_array, y_array, ycalc_array)

    kwargs : optional
        The kwargs arguments for the plotting of each data. It depends on mode.
        If mode = 'line', kwargs in ('xy_kwargs',).
        If mode = 'fit', kwargs in ('xy_kwargs', 'xycalc_kwargs', 'xydiff_kwargs', 'xyzero_kwargs',
        'fill_kwargs', 'yzero').

    mode : str
        The plotting mode. Currently support 'line', 'fit'.

    ax : Axes
        The axes to visualize the data. If None, use current axes. (Not used for CLI)

    normal : bool
        If True, the second and the following rows in data will be normalized by (max - min). Else, do nothing.

    stack : bool
        If True, the second and the third rows will be shifted so that there will be a gap between data (
        waterfall plot). Else, the data will be plotted without shifting (comparison plot).

    gap : float
        The gap between the adjacent curves. It is defined by the nearest points in vertical direction.

    texts : an iterable of str
        The texts to annotate the curves. It has the same order as the curves.

    text_xy : tuple
        The tuple of x and y position of the annotation in data coordinates. If None, use the default in the
        'tools.auto_text'.

    label : str
        The label type used in automatic labeling. Acceptable types are listed in 'tools._LABELS'. If None,
        the label will be guessed according to the suffix of the fist file in the list.

    minor_tick : int
        How many parts that the minor ticks separate the space between the two adjacent major ticks. Default 2.
        If None, no minor ticks.

    legends : a list of str
        The legend labels for the curves.

    colors : an iterable of colors
        The color of the plots. If None, use default color cycle in rc.

    show_fig : bool
        If True, the figure will be pop out and shown. Else, stay in the cache.

    Returns
    -------
    ax : Axes
        The axes with the plot.
    """
    if len(data_files) == 0:
        raise ValueError("No input file.")
    dataset = (io.load_array(_) for _ in data_files)
    if label is None:
        label = PurePath(data_files[0]).suffix.replace('.', '')
    ax = vis.waterfall(
        dataset, ax=ax, mode=mode, normal=normal, stack=stack, gap=gap, texts=texts, text_xy=text_xy,
        label=label, minor_tick=minor_tick, legends=legends, colors=colors, **kwargs
    )
    if show_fig:
        plt.show()
    return ax


def visualize(
    data_file: str, ax: Axes = None, mode: str = "line", normal: bool = False,
    text: str = None, text_xy: tuple = None, label: str = None,
    minor_tick: int = 2, legends: tp.List[str] = None, color: tp.Iterable = None,
    show_fig: bool = True, **kwargs
) -> Axes:
    """Visualize the data in a single data file.

    The data must be multi-columns matrix in a txt file. A header can be included in the file and it won't be
    read. The visualization has different modes for different kinds of plotting. Currently, it supports

    'line' mode
        The single curve of x and y data.

    'fit' mode
        One curve for data, one curve for the fit and a difference curve below it.

    Parameters
    ----------
    data_file : file path
        The file path. The file contains column data. The required format depends on the 'mode'.
        If mode = 'line', data = (x_array, y_array)
        If mode = 'fit', data = (x_array, y_array, ycalc_array)

    kwargs : optional
        The kwargs arguments for the plotting of each data. It depends on mode.
        If mode = 'line', kwargs in ('xy_kwargs',).
        If mode = 'fit', kwargs in ('xy_kwargs', 'xycalc_kwargs', 'xydiff_kwargs', 'xyzero_kwargs',
        'fill_kwargs', 'yzero').

    mode : str
        The plotting mode.

    ax : Axes
        The axes to visualize the data. If None, use current axes. (Not used for CLI)

    normal : bool
        If True, the second and the following rows in data will be normalized by (max - min). Else, do nothing.
        Defulat False.

    text : str
        The text to annotate the curve.

    text_xy : tuple
        The tuple of x and y position of the annotation in data coordinates. If None, use the default in the
        'tools.auto_text'.

    label : str
        The label type used in automatic labeling. Acceptable types are listed in 'tools._LABELS'

    minor_tick : int
        How many parts that the minor ticks separate the space between the two adjacent major ticks. Default 2.
        If None, no minor ticks.

    legends : list
        The legend label for the curve.

    color : an iterable of colors
        The color of the plots. If None, use default color cycle in rc.

    show_fig : bool
        If True, the figure will be pop out and shown. Else, stay in the cache.

    Returns
    -------
    ax : Axes
        The axes with the plot.
    """
    data = io.load_array(data_file)
    if label is None:
        label = PurePath(data_file).suffix.replace('.', '')
    ax = vis.visualize(
        data, ax=ax, mode=mode, normal=normal, text=text, text_xy=text_xy, label=label,
        minor_tick=minor_tick, legends=legends, color=color, **kwargs)
    if show_fig:
        plt.show()
    return ax



from collections import namedtuple

A = r"$\mathrm{\AA}$"
INV_A = r"$\mathrm{\AA}^{-1}$"
INV_SQ_A = r"$\mathrm{\AA}^{-2}$"
INV_NM = r"$nm^{-1}$"
MM = "mm"
DEG = r"deg"
RAD = r"rad"
ARB = "a. u."

LABELS = namedtuple("LABELS", ["xy", "chi", "iq", "sq", "fq", "gr", "fgr"])
LABELS.xy = (r"Q ($\mathrm{\AA}^{-1}$)", r"I (A. U.)")
LABELS.chi = (r"Q ($\mathrm{\AA}^{-1}$)", r"I (A. U.)")
LABELS.iq = (r"Q ($\mathrm{\AA}^{-1}$)", r"I (A. U.)")
LABELS.sq = (r"Q ($\mathrm{\AA}^{-1}$)", r"S")
LABELS.fq = (r"Q ($\mathrm{\AA}^{-1}$)", r"F ($\mathrm{\AA}^{-1}$)")
LABELS.gr = (r"r ($\mathrm{\AA}$)", r"G ($\mathrm{\AA}^{-2}$)")
LABELS.fgr = (r"r ($\mathrm{\AA}$)", r"G ($\mathrm{\AA}^{-2}$)")

MAP_PYFAI_TO_MPL = {
    "q_nm^-1": INV_NM,
    "q_A^-1": INV_A,
    "2th_deg": DEG,
    "2th_rad": RAD,
    "r_mm": MM
}

MAP_PYFAI_TO_XNAME = {
    "q_nm^-1": "Q",
    "q_A^-1": "Q",
    "2th_deg": "twotheta",
    "2th_rad": "twotheta",
    "r_mm": "r"
}

MAP_PYFAI_TO_YNAME = {
    "q_nm^-1": "I",
    "q_A^-1": "I",
    "2th_deg": "I",
    "2th_rad": "I",
    "r_mm": "I"
}

MAP_PYFAI_TO_DATAFORMAT = {
    "q_nm^-1": "Qnm",
    "q_A^-1": "QA",
    "2th_deg": "twotheta",
    "2th_rad": "twotheta",
}



from pathlib import Path
from tempfile import TemporaryDirectory

import matplotlib.pyplot as plt

from pdfstream.utils.jupyter import savefig_factory, FigExporter


def test_savefig_factory():
    plt.figure()
    with TemporaryDirectory() as d:
        figure_forlder = Path(d) / "figures"
        savefig = savefig_factory(figure_dir=figure_forlder)
        fname = 'test.png'
        savefig(fname)
        figure_path = figure_forlder / fname
        assert figure_path.exists()
    plt.clf()


def test_FigExporter():
    plt.figure()
    with TemporaryDirectory() as d:
        figure_dir = Path(d) / "test"
        exporter = FigExporter(str(figure_dir))
        exporter.update(dpi=40)
        exporter("test.svg")
        exporter.latex()
        target = figure_dir / "test.svg"
        assert target.is_file()
    plt.clf()



import pdfstream.utils.wrappers as wrappers


def test_report_stage(capsys):
    def func():
        return

    wrapped_func = wrappers.report_stage(func)
    wrapped_func()
    captured = capsys.readouterr()
    assert captured.out == "start func\nfinish func\n"


def test_take_in_namespace():
    def func(arg0, **kwargs):
        return locals()

    wrapped_func = wrappers.take_in_namespace(func)
    assert wrapped_func({'arg0': 0, 'arg1': 1}) == {'arg0': 0, 'arg1': 1}



from configparser import ConfigParser

from bluesky.callbacks.core import CallbackBase
from databroker.core import BlueskyRun


class AnalyzerConfig(ConfigParser):
    """The base class of configuration of analyzers."""

    def read_run(self, run: BlueskyRun, source="<BlueskyRun>"):
        """Read the configuration from the analysis result in a bluesky run."""
        # see schemas for the key of configuration
        config_dct = run.metadata["start"]["an_config"]
        return self.read_dict(config_dct, source=source)


class Analyzer(CallbackBase):
    """The base class of analyzers."""

    def analyze(self, run: BlueskyRun):
        """Analyze the data in a bluesky run."""
        for name, doc in run.canonical(fill="yes"):
            # inject the original_db
            if name == "start":
                doc = doc.to_dict()
                doc["original_db"] = run.catalog_object.name
            self.__call__(name, doc)



import datetime
import re
import string
from collections import defaultdict
from pathlib import Path


class PartialFormatter(string.Formatter):
    def get_field(self, field_name, args, kwargs):
        # Handle a key not found
        # Expects the return value to be a tuple of (obj,used_key)
        try:
            val = super(PartialFormatter, self).get_field(
                field_name, args, kwargs
            )
            # Python 3, 'super().get_field(field_name, args, kwargs)' works
        except (KeyError, AttributeError):
            val = "{" + field_name + "}", field_name
        return val

    def format_field(self, value, spec):
        # handle an invalid format
        if value is None:
            return spec
        try:
            return super(PartialFormatter, self).format_field(value, spec)
        except ValueError:
            return value[:-1] + ":" + spec + value[-1]


class CleanFormatter(string.Formatter):
    def get_field(self, field_name, args, kwargs):
        # Handle a key not found
        try:
            val = super(CleanFormatter, self).get_field(
                field_name, args, kwargs
            )
            # Python 3, 'super().get_field(field_name, args, kwargs)' works
        except (KeyError, AttributeError):
            val = "", field_name
        return val

    def format_field(self, value, spec):
        # handle an invalid format
        if value is None:
            return spec
        try:
            return super(CleanFormatter, self).format_field(value, spec)
        except ValueError:
            return ""


cfmt = CleanFormatter()
pfmt = PartialFormatter()


def clean_template(template, removals=None, cfmt=cfmt):
    if removals is None:
        removals = ["temp", "dx", "dy"]
    # this will essentially replace any nonexistent keys (field names) with ''
    d = cfmt.format(template, defaultdict(str))

    for r in removals:
        d = d.replace("[{}_]".format(r), "")
    z = re.sub(r"_+", "_", d)
    z = z.replace("_.", ".")
    z = re.sub(r"[\[\]\(\)\']", "", z)
    f = Path(z).as_posix()
    f = f.replace("/_", "/")
    return f


# Replaces each parameter name within folder_tag_list with the
# value of that parameter.
def get_filename_prefix(folder_tag_list, md):
    result = ""
    for tag in folder_tag_list:
        if isinstance(tag, str):
            raw_addition = md.get(tag, "")
        else:
            sub_md = md.copy()
            for item in tag:
                if isinstance(sub_md, dict):
                    sub_md = sub_md.get(item, "")
            raw_addition = sub_md
        if type(raw_addition) != float:
            addition = str(raw_addition)
        else:
            if (raw_addition - int(raw_addition)) == 0:
                addition = str(int(raw_addition))
            else:
                addition = str(raw_addition).replace(".", "-")
        result += addition + "/"
    return result


def render(string, formatter=pfmt, **kwargs):
    md = kwargs.get("raw_start", "")
    filename_prefix = ""
    if md != "":
        folder_tag_list = md.get("folder_tag_list", ["sample_name"])
        filename_prefix = get_filename_prefix(folder_tag_list, md)
        if re.fullmatch(r"/+", filename_prefix):
            filename_prefix = md.get("sample_name")
    formatted_string = formatter.format(
        string, folder_prefix=filename_prefix, **kwargs
    )
    return formatted_string


def render2(string: str, md: dict, formatter=pfmt):
    md = md.copy()
    md["time"] = timestampstr(md["time"])
    folder_tag_list = md.get("folder_tag_list", ["sample_name"])
    filename_prefix = get_filename_prefix(folder_tag_list, md)
    if re.fullmatch(r"/+", filename_prefix):
        filename_prefix = md.get("sample_name")
    formatted_string = formatter.format(
        string, folder_prefix=filename_prefix, **md
    )
    return formatted_string


def render_and_clean(string, md, formatter=pfmt):
    """Render and clean the file name."""
    f = clean_template(render2(string, md, formatter=formatter))
    print("saving file at {}".format(f))
    return f


def timestampstr(timestamp):
    """Convert timestamp to strftime formate """
    timestring = datetime.datetime.fromtimestamp(float(timestamp)).strftime(
        "%Y%m%d-%H%M%S"
    )
    return timestring


class SpecialStr(str):
    """A special string with a formatter that ignores the key error."""
    CF = CleanFormatter()

    def format(self, *args: object, **kwargs: object) -> str:
        return self.CF.format(self, *args, **kwargs)



import pytest

import pdfstream.callbacks.from_descriptor as mod

DCT0 = {'camera_image': {'dtype': 'number',
                         'shape': [512, 512],
                         'source': 'PV:...'}}

DCT1 = {'camera_image': {'dtype': 'array',
                         'shape': [],
                         'source': 'PV:...'}}

DCT2 = {'camera_image': {'dtype': 'number',
                         'shape': [512],
                         'source': 'PV:...'}}

DCT3 = {'camera_image': {'dtype': 'boolean',
                         'shape': [512, 512],
                         'source': 'PV:...'}}


@pytest.mark.parametrize(
    "dct, expect",
    [
        pytest.param(DCT1, 'camera_image'),
        pytest.param(DCT1, 'camera_image'),
        pytest.param(DCT2, 'camera_image', marks=pytest.mark.xfail),
        pytest.param(DCT3, 'camera_image', marks=pytest.mark.xfail)
    ]
)
def test_find_one_2darray(dct, expect):
    outcome = mod.find_one_array(dct)
    assert outcome == expect



from configparser import ConfigParser, Error
from pathlib import Path

from bluesky.callbacks import CallbackBase
from bluesky.callbacks.core import make_class_safe
from bluesky.callbacks.zmq import RemoteDispatcher

from pdfstream.io import server_message
from pdfstream.vend.qt_kicker import install_qt_kicker


class ServerConfig(ConfigParser):
    """The configuration for the server."""

    @property
    def host(self):
        return self.get("LISTEN TO", "host")

    @property
    def address(self):
        return self.host, self.port

    @property
    def port(self):
        return self.getint("LISTEN TO", "port")

    @property
    def prefix(self):
        return self.get("LISTEN TO", "prefix", fallback="").encode()


class BaseServer(RemoteDispatcher):
    """The basic server class."""

    def start(self):
        try:
            server_message("Server is started.")
            super(BaseServer, self).start()
        except KeyboardInterrupt:
            server_message("Server is terminated.")

    def install_qt_kicker(self):
        install_qt_kicker(self.loop)


def find_cfg_file(directory: Path, name: str) -> str:
    """Find the configuration file by matching the value of BASIC section name paramter to the name variable."""
    ini_files = []
    for filename in directory.glob("*.ini"):
        try:
            config = ConfigParser()
            ini_files.extend(config.read(str(filename)))
            if name == config["BASIC"]["name"]:
                return str(filename)
        except Error:
            continue
    raise FileNotFoundError(
        "\n".join(
            [
                "These are the ini_files found in the {}:".format(str(directory)),
                "\n".join(ini_files),
                "No .ini file satisfies that parameter 'name = {}' in section [BASIC]".format(name)
            ]
        )
    )


@make_class_safe
class StartStopCallback(CallbackBase):
    """Print the time for analysis"""

    def __init__(self):
        super(StartStopCallback, self).__init__()

    def start(self, doc):
        server_message("Receive the start of run {}".format(doc["uid"]))
        super(StartStopCallback, self).start(doc)

    def stop(self, doc):
        server_message("Receive the stop of run {}".format(doc.get("run_start", "")))
        super(StartStopCallback, self).stop(doc)



import fire

import pdfstream.cli as cli
from pdfstream.servers import ServerNames, CONFIG_DIR

try:
    import diffpy.pdfgetx

    PDFGETX_AVAILABLE = True
    del diffpy.pdfgetx
except ImportError:
    PDFGETX_AVAILABLE = False

COMMANDS = {
    'average': cli.average,
    'integrate': cli.integrate,
    'waterfall': cli.waterfall,
    'visualize': cli.visualize
}

SERVERS = {}

if PDFGETX_AVAILABLE:
    import pdfstream.transformation.cli

    import pdfstream.servers.xpd_server as xpd_server
    import pdfstream.servers.lsq_server as lsq_server

    COMMANDS.update({'transform': pdfstream.transformation.cli.transform})
    SERVERS.update({ServerNames.xpd: xpd_server.make_and_run})
    SERVERS.update({ServerNames.lsq: lsq_server.make_and_run})


def main():
    """The CLI entry point. Run google-fire on the name - function mapping."""
    fire.Fire(COMMANDS)


def run_server():
    """The CLI entry point. Run the server."""
    fire.Fire(SERVERS)


def print_server_config_dir():
    """The CLI entry point. Print the configuration directory for servers."""
    fire.Fire(lambda: print(str(CONFIG_DIR)))



import databroker
import pytest

import pdfstream.analyzers.base as mod
from pdfstream.callbacks.composer import gen_stream


@pytest.fixture(scope="function")
def db_with_fake_an():
    """A database that has a fake analysis run."""
    db = databroker.v2.temp()
    for name, doc in gen_stream([], {"an_config": {"SECTION": {"key": "value"}}}):
        db.v1.insert(name, doc)
    return db


def test_AnalyzerConfig(db_with_fake_an):
    db = db_with_fake_an
    config = mod.AnalyzerConfig()
    config.read_run(db[-1])
    assert config.sections() == ["SECTION"]
    assert config["SECTION"]["key"] == "value"


def test_Analyzer(db_with_fake_an):
    db = db_with_fake_an
    analyzer = mod.Analyzer()
    analyzer.analyze(db[-1])



import typing as tp

from pdfstream.errors import ValueNotFoundError


def find_one_array(data_keys: dict) -> str:
    """Fina one array in the data."""
    for key, value in data_keys.items():
        if value["dtype"] == "array":
            return key
    raise ValueNotFoundError("Array not found in {}".format(data_keys.keys()))


def yield_1d_array(data_keys: dict) -> tp.Generator[str, None, None]:
    """Fina one array in the data."""
    for key, value in data_keys.items():
        if value["dtype"] == "array" and len(value["shape"]) == 1:
            yield key


def find_one_image(descriptor: tp.Dict[str, dict]) -> str:
    """Find the key of image data in descriptor."""
    for key, dct in descriptor["data_keys"].items():
        if dct["dtype"] == "array" and _is_image(dct["shape"]):
            return key
    raise ValueNotFoundError("No image array found in descriptor (uid: {})".format(descriptor["uid"]))


def _is_image(shape: list) -> bool:
    """Whether the shape has 2 dimension that has rank larger than 1."""
    return sum([1 for s in shape if s > 1]) == 2



"""Some useful functions and classes when using jupyter lab."""
from pathlib import Path

import matplotlib.pyplot as plt
from matplotlib.figure import Figure

LATEX_REF = (
    "\\begin{{figure}}[hptb]\n"
    "\\includegraphics[width=\\columnwidth]{{{}}}\n"
    "\\caption{{{}}}\n"
    "\\label{{fig:{}}}\n"
    "\\end{{figure}}\n"
)


def savefig_factory(figure_dir):
    """
    Create function 'savefig' to save the current figure in the directory.

    Parameters
    ----------
    figure_dir
        The directory to save the figure.

    Returns
    -------
    save_fig
        A function to save current figure.
    """
    figure_dir = Path(figure_dir)
    if not figure_dir.is_dir():
        figure_dir.mkdir()

    def savefig(fname: str,
                dpi=240,
                facecolor=None,
                edgecolor=None,
                orientation='portrait',
                fmt='pdf',
                transparent=True,
                bbox_inches=None,
                pad_inches=0.1,
                metadata=None,
                latex=True,
                caption=""):
        """
        The output formats available depend on the backend being used.
        The latex reference will be copied to clipboard.

        Parameters
        ----------

        fname : str or PathLike or file-like object
            A path, or a Python file-like object, or
            possibly some backend-dependent object such as
            `matplotlib.backends.backend_pdf.PdfPages`.

            If *format* is not set, then the output format is inferred from
            the extension of *fname*, if any, and from :rc:`savefig.format`
            otherwise.  If *format* is set, it determines the output format.

            Hence, if *fname* is not a path or has no extension, remember to
            specify *format* to ensure that the correct backend is used.

        dpi : [ *None* | scalar > 0 | 'figure' ]
            The resolution in dots per inch.  If *None*, defaults to
            :rc:`savefig.dpi`.  If 'figure', uses the figure's dpi value.

        facecolor : color spec or None, optional
            The facecolor of the figure; if *None*, defaults to
            :rc:`savefig.facecolor`.

        edgecolor : color spec or None, optional
            The edgecolor of the figure; if *None*, defaults to
            :rc:`savefig.edgecolor`

        orientation : {'landscape', 'portrait'}
            Currently only supported by the postscript backend.

        fmt : str
            The file format, e.g. 'png', 'pdf', 'svg', ... The behavior when
            this is unset is documented under *fname*.

        transparent : bool
            If *True*, the axes patches will all be transparent; the
            figure patch will also be transparent unless facecolor
            and/or edgecolor are specified via kwargs.
            This is useful, for example, for displaying
            a plot on top of a colored background on a web page.  The
            transparency of these patches will be restored to their
            original values upon exit of this function.

        bbox_inches : str or `~matplotlib.transforms.Bbox`, optional
            Bbox in inches. Only the given portion of the figure is
            saved. If 'tight', try to figure out the tight bbox of
            the figure. If None, use savefig.bbox

        pad_inches : scalar, optional
            Amount of padding around the figure when bbox_inches is
            'tight'. If None, use savefig.pad_inches

        metadata : dict, optional
            Key/value pairs to store in the image metadata. The supported keys
            and defaults depend on the image format and backend:

            - 'png' with Agg backend: See the parameter ``metadata`` of
              `~.FigureCanvasAgg.print_png`.
            - 'pdf' with pdf backend: See the parameter ``metadata`` of
              `~.backend_pdf.PdfPages`.
            - 'eps' and 'ps' with PS backend: Only 'Creator' is supported.

        latex : bool, optional
            Whether to copy the latex reference into clipboard. The template is in the LATEX_REF.

        caption : str, optional
            The caption for the figure in the latex.
        """
        fname = figure_dir.joinpath(fname)
        plt.savefig(
            fname,
            dpi=dpi,
            facecolor=facecolor,
            edgecolor=edgecolor,
            orientation=orientation,
            format=fmt,
            transparent=transparent,
            bbox_inches=bbox_inches,
            pad_inches=pad_inches,
            metadata=metadata
        )
        if latex:
            print(
                LATEX_REF.format(fname.name, caption, fname.stem)
            )
        return

    return savefig


class FigExporter(object):
    """A figure exporter based on matplotlib.pyplot.savefig.

    Examples
    --------
    >>> exporter = FigExporter("my_figure_dir", dpi=600, fmt="png")
    >>> exporter("my_figure_name.png")
    >>> exporter.latex()
    """

    def __init__(self, figure_dir: str = ".", **kwargs):
        """Initialize the object. kwargs are used to update the _config attribute."""
        self._figure_dir = Path(figure_dir)
        if not self._figure_dir.is_dir():
            self._figure_dir.mkdir()
        self._filepath = None
        self._config = kwargs

    def __call__(self, filename: str, fig: Figure = None, **kwargs):
        """Save figure to directory and update the _filepath attribute."""
        fig = fig if fig is not None else plt.gcf()
        config = self._config.copy()
        config.update(kwargs)
        path = self._figure_dir.joinpath(filename)
        fig.savefig(str(path), **config)
        self._filepath = path

    def update(self, **kwargs):
        """Update the _config attribute."""
        self._config.update(kwargs)

    def latex(self, caption: str = ""):
        """Print the latex string of the reference of the figure in the _filepath attribute."""
        print(
            LATEX_REF.format(self._filepath.name, caption, self._filepath.stem)
        )



"""API to use pdfgetter to transform xrd data to pdf."""
from collections import OrderedDict

import matplotlib.pyplot as plt
from diffpy.pdfgetx import PDFGetter, PDFConfig
from matplotlib.gridspec import GridSpec
from numpy import ndarray

__all__ = [
    'make_pdfgetter',
    'use_pdfgetter',
    'visualize'
]


def make_pdfgetter(pdfconfig: PDFConfig, user_config: dict = None) -> PDFGetter:
    """Make the pdfgetter."""
    if user_config is not None:
        pdfconfig.update(**user_config)
    pdfgetter = PDFGetter(pdfconfig)
    return pdfgetter


def use_pdfgetter(chi: ndarray, pdfgetter: PDFGetter) -> PDFGetter:
    """Transform the data using the pdfgetter."""
    if len(chi.shape) != 2:
        raise ValueError("Wrong data shape in chi: {}".format(chi.shape))
    if chi.shape[0] != 2:
        raise ValueError("Wrong number of rows in chi: {}".format(chi.shape[0]))
    pdfgetter(chi[0], chi[1])
    return pdfgetter


def visualize(pdfgetter: PDFGetter, plot_setting: dict = None):
    """Visualize data in pdfgetter."""
    if plot_setting is None:
        plot_setting = {}
    # get data
    dct = OrderedDict()
    for attr in ("iq", "sq", "fq", "gr"):
        tup = getattr(pdfgetter, attr)
        if len(tup) > 0:
            dct[attr] = tup
    # plot data
    figsize = plt.rcParams['figure.figsize']
    fig = plt.figure(
        figsize=(figsize[0], figsize[1] / 2. * len(dct)),
        tight_layout=False
    )
    grids = GridSpec(len(dct), 1, hspace=0.5)
    labels = {
        "iq": (r"Q ($\AA^{-1}$)", r"I (A. U.)"),
        "sq": (r"Q ($\AA^{-1}$)", r"S"),
        "fq": (r"Q ($\AA^{-1}$)", r"F ($\AA^{-1}$)"),
        "gr": (r"r ($\AA$)", r"G ($\AA^{-2}$)")
    }
    for grid, (dtype, tup) in zip(grids, dct.items()):
        x, y = tup
        xlabel, ylabel = labels.get(dtype)
        ax = fig.add_subplot(grid)
        ax.plot(x, y, **plot_setting)
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
    plt.show(block=False)
    return



"""Decorators API."""
from functools import wraps
from typing import Callable


def report_stage(func: Callable):
    """Add 'start' at the beginning and 'finish' at the end."""

    @wraps(func)
    def _func(*args, **kwargs):
        print(f"start {func.__name__}")
        returns = func(*args, **kwargs)
        print(f"finish {func.__name__}")
        return returns

    return _func


def take_in_namespace(func: Callable):
    """Make function take a dictionary and return the name space. Flatten the kwargs in name space."""

    @wraps(func)
    def _func(namespace: dict, **kwargs):
        # pop out not used kwargs and unpack and put back to dictionary
        namespace.update(kwargs)
        namespace = func(**namespace)
        namespace.update(namespace.pop('kwargs', {}))
        return namespace

    return _func



import typing as tp
from configparser import ConfigParser, Error
from pathlib import Path

import databroker
import event_model
import matplotlib.pyplot as plt
import numpy as np
from bluesky.callbacks.stream import LiveDispatcher
from databroker.v2 import Broker
from event_model import RunRouter
from matplotlib.gridspec import GridSpec
from pyFAI.azimuthalIntegrator import AzimuthalIntegrator
from suitcase.csv import Serializer as CSVSerializer
from suitcase.json_metadata import Serializer as JsonSerializer
from suitcase.tiff_series import Serializer as TiffSerializer

import pdfstream
import pdfstream.callbacks.from_descriptor as from_desc
import pdfstream.callbacks.from_event as from_event
import pdfstream.callbacks.from_start as from_start
import pdfstream.integration.tools as integ
from pdfstream.callbacks.basic import MyLiveImage, LiveMaskedImage, LiveWaterfall, DataFrameExporter, \
    SmartScalarPlot, StackedNumpyTextExporter, NumpyExporter
from pdfstream.units import LABELS
from pdfstream.vend.formatters import SpecialStr

try:
    from diffpy.pdfgetx import PDFConfig, PDFGetter
except ImportError:
    pass


def no_need_to_refresh_db(db: tp.Union[None, Broker], name: tp.Union[None, str]):
    """If there is no need to refresh the db, return True. Used in configparser."""
    return (db is not None and name == db.name) or (db is None and name is None)


class BasicAnalysisConfig(ConfigParser):
    """The basic configuration that is shared by analysis and calibration."""

    def __init__(self, *args, **kwargs):
        super(BasicAnalysisConfig, self).__init__(*args, **kwargs)
        self._raw_db = None

    @property
    def raw_db(self) -> tp.Union[None, Broker]:
        name = self.get("DATABASE", "raw_db", fallback=None)
        if no_need_to_refresh_db(self._raw_db, name):
            pass
        # refresh the db
        elif name is None:
            self._raw_db = None
        elif name == "temp":
            print("Warning: a temporary db is created for raw db. It will be destroy at the end of the session.")
            self._raw_db = databroker.v2.temp()
        else:
            self._raw_db = databroker.catalog[name]
        return self._raw_db

    @raw_db.setter
    def raw_db(self, db: Broker):
        section_name = "DATABASE"
        db_key = "raw_db"
        if section_name not in self.sections():
            self.add_section(section_name)
        self.set(section_name, db_key, db.name)
        self._raw_db = db

    @property
    def dark_identifier(self):
        return self.get("METADATA", "dk_identifier")

    @property
    def dk_id_key(self):
        return self.get("METADATA", "dk_id_key")

    @property
    def composition_key(self):
        return self.get("METADATA", "composition_key")

    @property
    def wavelength_key(self):
        return self.get("METADATA", "wavelength_key")

    def to_dict(self):
        """Convert the configuration to a dictionary."""
        return {s: dict(self.items(s)) for s in self.sections()}


class AnalysisConfig(BasicAnalysisConfig):
    """The configuration for analysis callbacks."""

    @property
    def calibration_md_key(self):
        return self.get("METADATA", "calibration_md_key")

    @property
    def bkgd_sample_name_key(self):
        return self.get("METADATA", "bkgd_sample_name_key")

    @property
    def sample_name_key(self):
        return self.get("METADATA", "sample_name_key")

    @property
    def mask_setting(self):
        section = self["MASK SETTING"]
        return {
            "alpha": section.getfloat("alpha"),
            "edge": section.getint("edge"),
            "lower_thresh": section.getfloat("lower_thresh"),
            "upper_thresh": section.getfloat("upper_thresh")
        }

    @property
    def integ_setting(self):
        section = self["INTEGRATION SETTING"]
        return {
            "npt": section.getint("npt"),
            "correctSolidAngle": section.getboolean("correctSolidAngle"),
            "polarization_factor": section.getfloat("polarization_factor"),
            "method": section.get("method"),
            "normalization_factor": section.getfloat("normalization_factor"),
            "unit": "q_A^-1"
        }

    @property
    def trans_setting(self):
        section = self["TRANSFORMATION SETTING"]
        return {
            "rpoly": section.getfloat("rpoly"),
            "qmaxinst": section.getfloat("qmaxinst"),
            "qmin": section.getfloat("qmin"),
            "qmax": section.getfloat("qmax")
        }

    @property
    def grid_config(self):
        section = self["PDF SETTING"]
        return {
            "rmin": section.getfloat("rmin"),
            "rmax": section.getfloat("rmax"),
            "rstep": section.getfloat("rstep")
        }


class AnalysisStream(LiveDispatcher):
    """The secondary stream for data analysis.

    It inject the configuration into start document and emit processed data to the subscribers.
    """

    def __init__(self, config: AnalysisConfig):
        super(AnalysisStream, self).__init__()
        self.config = config
        self.cache = {}
        self.db = config.raw_db

    def start(self, doc, _md=None):
        self.cache = dict()
        self.cache["start"] = doc
        self.cache["ai"] = from_start.query_ai(
            doc,
            calibration_md_key=self.config.calibration_md_key
        )
        self.cache["bt_info"] = from_start.query_bt_info(
            doc,
            composition_key=self.config.composition_key,
            wavelength_key=self.config.wavelength_key,
            default_composition="Ni"
        )
        self.cache["indeps"] = from_start.get_indeps(doc, exclude={"time"})
        new_start = dict(**doc, an_config=self.config.to_dict(), pdfstream_version=pdfstream.__version__)
        super(AnalysisStream, self).start(new_start)

    def event_page(self, doc):
        for event_doc in event_model.unpack_event_page(doc):
            self.event(event_doc)

    def descriptor(self, doc):
        self.cache["det_name"] = from_desc.find_one_image(doc)
        self.cache["dk_img"] = from_start.query_dk_img(
            self.cache["start"],
            db=self.db,
            dk_id_key=self.config.dk_id_key,
            det_name=self.cache["det_name"]
        )
        self.cache["dk_sub_bg_img"] = from_start.query_bg_img(
            self.cache["start"],
            bkgd_sample_name_key=self.config.bkgd_sample_name_key,
            sample_name_key=self.config.sample_name_key,
            det_name=self.cache["det_name"],
            dk_id_key=self.config.dk_id_key,
            db=self.db
        )
        super(AnalysisStream, self).descriptor(doc)

    def event(self, doc, _md=None):
        # the raw image in the data
        raw_img = from_event.get_image_from_event(doc, det_name=self.cache["det_name"])
        # the independent variables in the data
        indep_data = {key: doc["data"][key] for key in self.cache["indeps"]}
        # process the data output a dictionary
        an_data = process(
            raw_img=raw_img,
            ai=self.cache["ai"],
            dk_img=self.cache["dk_img"],
            dk_sub_bg_img=self.cache["dk_sub_bg_img"],
            integ_setting=self.config.integ_setting,
            mask_setting=self.config.mask_setting,
            pdfgetx_setting=dict(
                self.cache["bt_info"],
                **self.config.trans_setting,
                **self.config.grid_config
            )
        )
        # the final output data is a combination of the independent variables and processed data
        data = dict(**indep_data, **an_data)
        self.process_event(EventDoc(data=data, descriptor=doc["descriptor"]))

    def stop(self, doc, _md=None):
        super(AnalysisStream, self).stop(doc)


def process(
    *,
    raw_img: np.ndarray,
    ai: AzimuthalIntegrator,
    dk_img: np.ndarray = None,
    dk_sub_bg_img: np.ndarray = None,
    integ_setting: dict = None,
    mask_setting: dict = None,
    pdfgetx_setting: dict = None,
) -> dict:
    """The function to process the data from event."""
    data = dict()
    # dark subtraction
    if dk_img is None:
        dk_img = np.zeros_like(raw_img)
    dk_sub_img = np.subtract(raw_img, dk_img)
    data.update({"dk_sub_image": dk_sub_img})
    # background subtraction
    if dk_sub_bg_img is None:
        dk_sub_bg_img = np.zeros_like(dk_sub_img)
    bg_sub_img = np.subtract(dk_sub_img, dk_sub_bg_img)
    data.update({"bg_sub_image": bg_sub_img})
    # auto masking
    mask, _ = integ.auto_mask(bg_sub_img, ai, mask_setting=mask_setting)
    data.update({"mask": mask})
    # integration
    x, y = ai.integrate1d(bg_sub_img, mask=mask, **integ_setting)
    chi_max_ind = np.argmax(y)
    data.update({"chi_Q": x, "chi_I": y, "chi_max": y[chi_max_ind], "chi_argmax": x[chi_max_ind]})
    # transformation
    pdfconfig = PDFConfig(dataformat="QA", **pdfgetx_setting)
    pdfgetter = PDFGetter(pdfconfig)
    pdfgetter(x, y)
    iq, sq, fq, gr = pdfgetter.iq, pdfgetter.sq, pdfgetter.fq, pdfgetter.gr
    gr_max_ind = np.argmax(gr[1])
    data.update(
        {
            "iq_Q": iq[0], "iq_I": iq[1], "sq_Q": sq[0], "sq_S": sq[1], "fq_Q": fq[0], "fq_F": fq[1],
            "gr_r": gr[0], "gr_G": gr[1], "gr_max": gr[1][gr_max_ind], "gr_argmax": gr[0][gr_max_ind]
        }
    )
    return data


class EventDoc(dict):
    """A simplified event document for callbacks.

    It only contains two necessary key: data and descriptor. The data is the dictionary of data key and data
    value and the descriptor is the uid of the descriptor from the original event.
    """

    def __init__(self, data: dict, descriptor: str, **kwargs):
        super().__init__(data=data, descriptor=descriptor, **kwargs)

    @property
    def data(self):
        return self["data"]

    @data.setter
    def data(self, val: dict):
        self["data"] = val

    @property
    def descriptor(self):
        return self["descriptor"]

    @descriptor.setter
    def descriptor(self, val: dict):
        self["descriptor"] = val


class BasicExportConfig(ConfigParser):
    """Basic configuration that is shared by export and calibration."""

    @property
    def tiff_base(self):
        section = self["FILE SYSTEM"]
        dir_path = section.get("tiff_base")
        if not dir_path:
            raise Error("Missing tiff_base in configuration.")
        path = Path(dir_path).expanduser()
        return path

    @tiff_base.setter
    def tiff_base(self, value: str):
        self.set("FILE SYSTEM", "tiff_base", value)


class ExportConfig(BasicExportConfig):
    """The configuration of exporter."""
    @property
    def tiff_setting(self):
        """Settings for all tiff files."""
        section = self["TIFF SETTING"]
        if not section.getboolean("enable", fallback=True):
            return None
        return {"file_prefix": SpecialStr(section.get("file_prefix"))}

    @property
    def json_setting(self):
        """Settings for all json files of metadata."""
        section = self["JSON SETTING"]
        if not section.getboolean("enable", fallback=True):
            return None
        return {"file_prefix": SpecialStr(section.get("file_prefix"))}

    @property
    def csv_setting(self):
        """Settings for the csv file of the scalr data like temperature."""
        section = self["CSV SETTING"]
        if not section.getboolean("enable", fallback=True):
            return None
        return {"file_prefix": SpecialStr(section.get("file_prefix"))}

    @property
    def npy_setting(self):
        """Settings for the csv file of the processed data."""
        section = self["NPY SETTING"]
        if not section.getboolean("enable", fallback=True):
            return None
        return {"file_prefix": SpecialStr(section.get("file_prefix"))}

    @property
    def txt_setting(self):
        """Settings for txt file of processed 1d data."""
        section = self["NPY TXT SETTING"]
        if not section.getboolean("enable", fallback=True):
            return None
        return {"file_prefix": SpecialStr(section.get("file_prefix"))}

    @property
    def msk_npy_setting(self):
        """Settings for the npy file of the mask."""
        section = self["MASK NPY SETTING"]
        if not section.getboolean("enable", fallback=True):
            return None
        return {"file_prefix": SpecialStr(section.get("file_prefix"))}


class Exporter(RunRouter):
    """Export the processed data to file systems."""

    def __init__(self, config: ExportConfig):
        factory = ExporterFactory(config)
        super().__init__([factory])


class ExporterFactory:
    """The factory for the exporter run router."""

    def __init__(self, config: ExportConfig):
        self.config = config

    def __call__(self, name: str, doc: dict) -> tp.Tuple[list, list]:
        if name != "start":
            return [], []
        tiff_base = self.config.tiff_base
        tiff_base.mkdir(exist_ok=True, parents=True)
        callbacks = []
        if self.config.tiff_setting is not None:
            cb = TiffSerializer(
                str(tiff_base.joinpath("images")),
                **self.config.tiff_setting
            )
            callbacks.append(cb)
        if self.config.json_setting is not None:
            cb = JsonSerializer(
                str(tiff_base.joinpath("metadata")),
                **self.config.json_setting
            )
            callbacks.append(cb)
        if self.config.csv_setting is not None:
            cb = CSVSerializer(
                str(tiff_base.joinpath("scalar_data")),
                **self.config.csv_setting
            )
            callbacks.append(cb)
        if self.config.npy_setting is not None:
            cb = DataFrameExporter(
                str(tiff_base.joinpath("array_data")),
                **self.config.npy_setting
            )
            callbacks.append(cb)
        return callbacks, []


class ExporterXpdan(RunRouter):
    """Export the processed data to file systems in an xpdan style file structure."""

    def __init__(self, config: ExportConfig):
        factory = ExporterXpdanFactory(config)
        super().__init__([factory])


class ExporterXpdanFactory:
    """The factory that gives the callbacks to export analysis data to xpdan style directory structure."""

    def __init__(self, config: ExportConfig):
        self.config = config

    def __call__(self, name: str, doc: dict) -> tp.Tuple[list, list]:
        if name != "start":
            return [], []
        folder_name = doc.get("sample_name", "unknown_sample")
        data_folder = self.config.tiff_base.joinpath(folder_name)
        data_folder.mkdir(exist_ok=True, parents=True)
        callbacks = []
        if self.config.tiff_setting is not None:
            cb = TiffSerializer(
                str(data_folder.joinpath("dark_sub")),
                **self.config.tiff_setting
            )
            callbacks.append(cb)
        if self.config.json_setting is not None:
            cb = JsonSerializer(
                str(data_folder.joinpath("meta")),
                **self.config.json_setting
            )
            callbacks.append(cb)
        if self.config.csv_setting is not None:
            cb = CSVSerializer(
                str(data_folder.joinpath("scalar_data")),
                **self.config.csv_setting
            )
            callbacks.append(cb)
        if self.config.txt_setting is not None:
            cb = StackedNumpyTextExporter(
                str(data_folder.joinpath("integration")),
                **self.config.txt_setting,
                data_keys=["chi_Q", "chi_I"]
            )
            callbacks.append(cb)
            cb = StackedNumpyTextExporter(
                str(data_folder.joinpath("iq")),
                **self.config.txt_setting,
                data_keys=["iq_Q", "iq_I"]
            )
            callbacks.append(cb)
            cb = StackedNumpyTextExporter(
                str(data_folder.joinpath("sq")),
                **self.config.txt_setting,
                data_keys=["sq_Q", "sq_S"]
            )
            callbacks.append(cb)
            cb = StackedNumpyTextExporter(
                str(data_folder.joinpath("fq")),
                **self.config.txt_setting,
                data_keys=["fq_Q", "fq_F"]
            )
            callbacks.append(cb)
            cb = StackedNumpyTextExporter(
                str(data_folder.joinpath("pdf")),
                **self.config.txt_setting,
                data_keys=["gr_r", "gr_G"]
            )
            callbacks.append(cb)
        if self.config.msk_npy_setting is not None:
            cb = NumpyExporter(
                str(data_folder.joinpath("mask")),
                **self.config.msk_npy_setting,
                data_keys=["mask"]
            )
            callbacks.append(cb)
        return callbacks, []


class VisConfig(ConfigParser):
    """The configuration of visualization."""

    @property
    def vis_masked_image(self):
        section = self["VIS MASKED IMAGE"]
        if not section.getboolean("enable", fallback=True):
            return None
        return {
            "cmap": section.get("cmap", fallback="viridis")
        }

    @property
    def vis_bg_sub_image(self):
        section = self["VIS BG SUB IMAGE"]
        if not section.getboolean("enable", fallback=True):
            return None
        return {
            "cmap": section.get("cmap", fallback="viridis")
        }

    @property
    def vis_dk_sub_image(self):
        section = self["VIS DK SUB IMAGE"]
        if not section.getboolean("enable", fallback=True):
            return None
        return {"cmap": section.get("cmap", fallback="viridis")}

    @property
    def vis_chi(self):
        section = self["VIS CHI"]
        if not section.getboolean("enable", fallback=True):
            return None
        return {
            "xlabel": section.get("xlabel", fallback=LABELS.chi[0]),
            "ylabel": section.get("ylabel", fallback=LABELS.chi[1])
        }

    @property
    def vis_iq(self):
        section = self["VIS IQ"]
        if not section.getboolean("enable", fallback=True):
            return None
        return {
            "xlabel": section.get("xlabel", fallback=LABELS.iq[0]),
            "ylabel": section.get("ylabel", fallback=LABELS.iq[1])
        }

    @property
    def vis_sq(self):
        section = self["VIS SQ"]
        if not section.getboolean("enable", fallback=True):
            return None
        return {
            "xlabel": section.get("xlabel", fallback=LABELS.sq[0]),
            "ylabel": section.get("ylabel", fallback=LABELS.sq[1])
        }

    @property
    def vis_fq(self):
        section = self["VIS FQ"]
        if not section.getboolean("enable", fallback=True):
            return None
        return {
            "xlabel": section.get("xlabel", fallback=LABELS.fq[0]),
            "ylabel": section.get("ylabel", fallback=LABELS.fq[1])
        }

    @property
    def vis_gr(self):
        section = self["VIS GR"]
        if not section.getboolean("enable", fallback=True):
            return None
        return {
            "xlabel": section.get("xlabel", fallback=LABELS.gr[0]),
            "ylabel": section.get("ylabel", fallback=LABELS.gr[1])
        }

    @property
    def vis_gr_max(self):
        section = self["VIS GR MAX"]
        if not section.getboolean("enable", fallback=True):
            return None
        return {"ylabel": section.get("ylabel", fallback=LABELS.gr[1])}

    @property
    def vis_gr_argmax(self):
        section = self["VIS GR ARGMAX"]
        if not section.getboolean("enable", fallback=True):
            return None
        return {"ylabel": section.get("ylabel", fallback=LABELS.gr[0])}

    @property
    def vis_chi_max(self):
        section = self["VIS CHI MAX"]
        if not section.getboolean("enable", fallback=True):
            return None
        return {"ylabel": section.get("ylabel", fallback=LABELS.chi[1])}

    @property
    def vis_chi_argmax(self):
        section = self["VIS CHI ARGMAX"]
        if not section.getboolean("enable", fallback=True):
            return None
        return {"ylabel": section.get("ylabel", fallback=LABELS.chi[0])}


class Visualizer(RunRouter):
    """Visualize the analyzed data. It can be subscribed to a live dispatcher."""

    def __init__(self, config: VisConfig):
        self._factory = VisFactory(config)
        super(Visualizer, self).__init__([self._factory])

    def show_figs(self):
        """Show all the figures in the callbacks in the factory."""
        return self._factory.show()


class VisFactory:
    """The factory of visualization callbacks."""

    def __init__(self, config: VisConfig):
        self.config = config
        self.cb_lst = []
        # images
        if self.config.vis_dk_sub_image is not None:
            self.cb_lst.append(
                MyLiveImage("dk_sub_image", **self.config.vis_dk_sub_image)
            )
        if self.config.vis_bg_sub_image is not None:
            self.cb_lst.append(
                MyLiveImage("bg_sub_image", **self.config.vis_dk_sub_image)
            )
        if self.config.vis_masked_image is not None:
            self.cb_lst.append(
                LiveMaskedImage("bg_sub_image", "mask", **self.config.vis_masked_image)
            )
        # one dimensional array waterfall
        for xfield, yfield, vis_config in [
            ("chi_Q", "chi_I", self.config.vis_chi),
            ("iq_Q", "iq_I", self.config.vis_iq),
            ("sq_Q", "sq_S", self.config.vis_sq),
            ("fq_Q", "fq_F", self.config.vis_fq),
            ("gr_r", "gr_G", self.config.vis_gr)
        ]:
            if vis_config is not None:
                fig = plt.figure()
                self.cb_lst.append(
                    LiveWaterfall(xfield, yfield, ax=fig.add_subplot(111), **vis_config)
                )
                fig.show()

        # scalar data
        fields_and_configs = [
            ("chi_max", self.config.vis_chi_max),
            ("chi_argmax", self.config.vis_chi_argmax),
            ("gr_max", self.config.vis_gr_max),
            ("gr_argmax", self.config.vis_gr_argmax)
        ]
        # make a figure of stacking axes
        fig1 = plt.figure()
        axes1 = [fig1.add_subplot(grid) for grid in GridSpec(len(fields_and_configs), 1, hspace=0)]
        if len(axes1) > 1:
            for ax in axes1[:-1]:
                ax.get_xaxis().set_visible(False)
        # link the axes with callbacks
        for (field, vis_config), ax in zip(fields_and_configs, axes1):
            if vis_config is not None:
                self.cb_lst.append(
                    SmartScalarPlot(field, ax=ax, **vis_config, marker="o")
                )
        fig1.show()

    def __call__(self, name: str, doc: dict) -> tp.Tuple[list, list]:
        if name != "start":
            return [], []
        return self.cb_lst, []

    def show(self):
        for cb in self.cb_lst:
            try:
                cb.show()
            except AttributeError:
                pass
        return



import typing as tp

import numpy as np
from numpy import ndarray
from pyFAI import AzimuthalIntegrator

from pdfstream.integration.tools import bg_sub, vis_img, integrate, vis_chi, auto_mask


def get_chi(
    ai: AzimuthalIntegrator,
    img: ndarray,
    dk_img: ndarray = None,
    bg_img: ndarray = None,
    mask: ndarray = None,
    bg_scale: float = None,
    mask_setting: tp.Union[str, dict] = None,
    integ_setting: dict = None,
    img_setting: tp.Union[str, dict] = None,
    plot_setting: tp.Union[str, dict] = None,
) -> tp.Tuple[
    ndarray,
    ndarray,
    ndarray,
    ndarray,
    tp.Union[None, ndarray],
    dict,
    tp.Union[str, dict]
]:
    """Process the diffraction image to get I(Q).

    The image will be subtracted by the background image and then auto masked. The results will be
    binned on the azimuthal direction and the average value of the intensity in the bin and their
    corresponding Q will be returned. The I(Q) and background subtracted masked image will
    be visualized.

    Parameters
    ----------
    ai : AzimuthalIntegrator
        The AzimuthalIntegrator.

    img : ndarray
        The of the 2D array of the image.

    dk_img : ndarray
        The dark frame image. The image will be subtracted by it if provided.

    bg_img : ndarray
        The 2D array of the background image. If None, no background subtraction.

    mask : ndarray
        A mask provided by user. The auto generated mask will be multiplied by this mask.

    bg_scale : float
        The scale for background subtraction. If None, use 1.

    mask_setting : dict
        The auto final_mask setting.
        If None, use _AUTOMASK_SETTING in the tools module. To turn off the auto masking, use "OFF".

    integ_setting : dict
        The integration setting.
        If None, use _INTEG_SETTING in the tools module.

    img_setting : dict
        The user's modification to imshow kwargs except a special key 'z_score'. If None, use use empty dict.
        To turn off the imshow, use "OFF".

    plot_setting : dict
        The kwargs for the plot function. If None, use empty dict.

    Returns
    -------
    chi : ndarray
        The 2D array of integrated results. The first row is the Q and the second row is the I.

    bg_sub_img : ndarray
        The background subtracted image. If no background subtraction, it is the dk_sub_image.

    dk_sub_img : ndarray
        The dark subtracted image. If no dark subtraction, it is the input img.

    img : ndarray
        The input image.

    final_mask : ndarray or None
        The final_mask array. If no auto_masking, return None.

    _integ_setting: dict
        The integration setting used.

    _mask_setting : dict or str
        The auto masking setting.
    """
    if dk_img is not None:
        dk_sub_img = bg_sub(img, dk_img, bg_scale=1.)
    else:
        dk_sub_img = img
    if bg_img is not None:
        bg_sub_img = bg_sub(dk_sub_img, bg_img, bg_scale=bg_scale)
    else:
        bg_sub_img = dk_sub_img
    if mask_setting != "OFF":
        final_mask, _mask_setting = auto_mask(bg_sub_img, ai, user_mask=mask, mask_setting=mask_setting)
    elif mask is not None:
        final_mask, _mask_setting = mask, {}
    else:
        final_mask, _mask_setting = None, None
    if img_setting != "OFF":
        vis_img(bg_sub_img, final_mask, img_setting=img_setting)
    chi, _integ_setting = integrate(bg_sub_img, ai, mask=final_mask, integ_setting=integ_setting)
    if plot_setting != "OFF":
        vis_chi(chi, plot_setting=plot_setting, unit=_integ_setting.get('unit'))
    return chi, bg_sub_img, dk_sub_img, img, final_mask, _integ_setting, _mask_setting


def avg_imgs(imgs: tp.Iterable[ndarray], weights: tp.Iterable[float] = None) -> ndarray:
    """Average the 2D images.

    Parameters
    ----------
    imgs : ndarray
        The 2D array of diffraction images.

    weights : an iterable of floats
        The weights for the images. If None, images will not be weighted when averaged.

    Returns
    -------
    avg_img : ndarray
        The averaged 2D image array.
    """
    return np.average(np.stack(tuple(imgs), axis=0), axis=0, weights=weights)



from pathlib import Path

import databroker
import databroker.v2
import pytest
from pkg_resources import resource_filename

import pdfstream.servers.xpd_server as mod

fn = resource_filename("tests", "configs/xpd_server.ini")


def test_XPDServer(tmpdir):
    config = mod.XPDServerConfig()
    config.read(fn)
    config.tiff_base = str(tmpdir)
    config.calib_base = str(tmpdir)
    config.raw_db = databroker.v2.temp()
    config["FUNCTIONALITY"]["send_messages"] = "True"
    mod.XPDServer(config)


@pytest.mark.parametrize(
    "start_doc, expect_lst",
    [
        ({"dark_frame": True}, []),
        ({"is_calibration": True}, [mod.Calibration]),
        ({}, [mod.AnalysisStream])
    ]
)
def test_XPDFactory(start_doc, expect_lst):
    """Assert XPDFactory output the correct callback function list."""
    config = mod.XPDConfig()
    config.read(fn)
    factory = mod.XPDFactory(config)
    out_lst, _ = factory("start", start_doc)
    _, _ = factory("stop", {})
    for obj, cls in zip(out_lst, expect_lst):
        assert isinstance(obj, cls)


def test_XPDRouter(db_with_img_and_bg_img, tmpdir):
    raw_db = db_with_img_and_bg_img
    config = mod.XPDConfig()
    config.read(fn)
    config.tiff_base = str(tmpdir)
    config.calib_base = str(tmpdir)
    an_db = config.an_db
    cb = mod.XPDRouter(config)
    for name, doc in raw_db[-1].canonical(fill="yes", strict_order=True):
        cb(name, doc)
    tiff_base = Path(config.tiff_base)
    assert len(list(tiff_base.rglob("*.tiff"))) > 0
    assert len(list(tiff_base.rglob("*.json"))) > 0
    assert len(list(tiff_base.rglob("*.csv"))) > 0
    assert len(list(an_db)) > 0


def test_XPDRouter_with_xpdan_exporter(db_with_img_and_bg_img, tmpdir):
    raw_db = db_with_img_and_bg_img
    config = mod.XPDConfig()
    config.read(fn)
    config.tiff_base = str(tmpdir)
    config.calib_base = str(tmpdir)
    config["FUNCTIONALITY"]["export_files_in_xpdan_style"] = "True"
    config["FUNCTIONALITY"]["export_files"] = "False"
    cb = mod.XPDRouter(config)
    for name, doc in raw_db[-1].canonical(fill="yes", strict_order=True):
        cb(name, doc)
    tiff_base = Path(config.tiff_base)
    data_folder = tiff_base.joinpath("Ni")
    assert data_folder.is_dir()
    for dir_name in ("dark_sub", "integration", "meta", "mask", "iq", "sq", "fq", "pdf", "scalar_data"):
        assert data_folder.joinpath(dir_name).is_dir()
        assert len(list(data_folder.joinpath(dir_name).glob("*.*"))) > 0
    an_db = config.an_db
    assert len(list(an_db)) > 0



import subprocess
import typing as tp
from configparser import Error
from pathlib import Path

import event_model
import numpy as np
from bluesky.callbacks.core import CallbackBase
from tifffile import TiffWriter

import pdfstream.callbacks.from_descriptor as fd
import pdfstream.callbacks.from_event as fe
import pdfstream.callbacks.from_start as fs
import pdfstream.io as io
from .analysis import BasicAnalysisConfig, BasicExportConfig


class CalibrationConfig(BasicAnalysisConfig, BasicExportConfig):
    """The configuration of the calibration callbacks."""

    @property
    def calib_identifier(self):
        return self.get("METADATA", "calib_identifier")

    @property
    def default_calibrant(self):
        return self.get("CALIBRATION", "default_calibrant")

    @property
    def detector_key(self):
        return self.get("METADATA", "detector_key")

    @property
    def calibrant_key(self):
        return self.get("METADATA", "calibrant_key")

    @property
    def calib_base(self):
        dir_path = self.get("FILE SYSTEM", "calib_base")
        if not dir_path:
            raise Error("Missing calib_base in configuration.")
        path = Path(dir_path).expanduser()
        return path

    @calib_base.setter
    def calib_base(self, value: str):
        self.set("FILE SYSTEM", "calib_base", value)

    @property
    def calib_tiff_dir(self):
        dir_path = self.tiff_base.joinpath("calib")
        return dir_path

    @property
    def poni_file(self):
        return self.get("CALIBRATION", "poni_file")


class Calibration(CallbackBase):
    """Run the calibration in a gui and save the results which will be used by xpdacq."""

    def __init__(self, config: CalibrationConfig, *, test: bool = False):
        super(Calibration, self).__init__()
        self.config = config
        self.cache = dict()
        self.db = self.config.raw_db
        self.test = test

    def start(self, doc):
        super(Calibration, self).start(doc)
        self.cache = dict()
        self.cache["img_sum"] = 0
        self.cache["img_num"] = 0
        self.cache["start"] = doc
        self.cache["calib_info"] = fs.get_calib_info(
            doc,
            wavelength_key=self.config.wavelength_key,
            detector_key=self.config.detector_key,
            calibrant_key=self.config.calibrant_key
        )

    def descriptor(self, doc):
        super(Calibration, self).descriptor(doc)
        self.cache["det_name"] = fd.find_one_image(doc)
        self.cache["dk_img"] = fs.query_dk_img(
            self.cache["start"],
            db=self.db,
            dk_id_key=self.config.dk_id_key,
            det_name=self.cache["det_name"]
        )

    def event_page(self, doc):
        for event_doc in event_model.unpack_event_page(doc):
            self.event(event_doc)

    def event(self, doc):
        super(Calibration, self).event(doc)
        raw_img = fe.get_image_from_event(doc, self.cache["det_name"])
        self.cache["img_sum"] += raw_img
        self.cache["img_num"] += 1

    def stop(self, doc):
        super(Calibration, self).stop(doc)
        self.config.calib_base.mkdir(parents=True, exist_ok=True)
        self.config.calib_tiff_dir.mkdir(parents=True, exist_ok=True)
        poni_path = self.config.calib_base.joinpath(self.config.poni_file)
        tiff_path = self.config.calib_tiff_dir.joinpath("{}-calib.tiff".format(self.cache["start"]["uid"]))
        calc_image_and_save(
            self.cache["img_sum"],
            self.cache["img_num"],
            self.cache["dk_img"],
            tiff_path
        )
        run_calibration_gui(
            tiff_path,
            poni_path,
            **self.cache["calib_info"],
            test=self.test
        )
        self.cache = dict()


def calc_image_and_save(img_sum: np.ndarray, img_num: int, dk_img: tp.Union[None, np.ndarray],
                        tiff_path: tp.Union[str, Path]) -> None:
    """Average the image, subtract the dark image and export the image in a tiff file."""
    avg_img = img_sum / img_num
    if dk_img is not None:
        avg_img = avg_img - dk_img
    tw = TiffWriter(tiff_path)
    tw.write(avg_img)
    return


def run_calibration_gui(
    tiff_file: Path,
    poni_file: Path,
    *,
    wavelength: float = "",
    calibrant: str = "",
    detector: str = "",
    test: bool = False
) -> int:
    """Run the gui of calibration."""
    args = ["pyFAI-calib2", "--poni", str(poni_file)]
    if wavelength:
        args.extend(["--wavelength", wavelength])
    if calibrant:
        args.extend(["--calibrant", calibrant])
    if detector:
        args.extend(["--detector", detector])
    args.append(str(tiff_file))
    if test:
        return 0
    cp = subprocess.run(args, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    if cp.returncode != 0:
        io.server_message("Error in Calibration. See below:")
        print(r"$", " ".join(args))
        print(cp.stdout.decode())
        print(cp.stderr.decode())
    return cp.returncode



import matplotlib.pyplot as plt
import pytest

import pdfstream.visualization.main as vis


@pytest.mark.parametrize(
    'keys,kwargs', [
        (['Ni_gr', 'Ni_gr'], {'mode': 'line', 'legends': ['Ni0', 'Ni1'], 'label': 'gr'}),
        (['Ni_gr', 'Ni_gr'], {'mode': 'line', 'stack': False}),
        (['Ni_gr', 'Ni_gr'], {'mode': 'line', 'xy_kwargs': {'color': 'black'}, 'texts': ['Ni0', 'Ni1']}),
        (['Ni_gr', 'Ni_gr'], {'mode': 'line', 'colors': ['r', 'g']}),
        (['Ni_fgr', 'Ni_fgr'], {'mode': 'fit', 'texts': ['Ni0', 'Ni1'], 'label': 'gr'}),
        (['Ni_fgr', 'Ni_fgr'], {'mode': 'fit', 'stack': False}),
        (['Ni_fgr', 'Ni_fgr'], {'mode': 'fit', 'xy_kwargs': {'color': 'black'}}),
        (['Ni_fgr', 'Ni_fgr'], {'mode': 'fit', 'colors': ['r', 'g']})
    ]
)
def test_waterfall(test_data, keys, kwargs):
    plt.figure()
    vis.waterfall((test_data[k] for k in keys), **kwargs)
    plt.show(block=False)
    plt.close()


def test_waterfall_error(test_data):
    plt.figure()
    with pytest.raises(ValueError):
        vis.waterfall(test_data['Ni_gr'], mode="unknown")


@pytest.mark.parametrize(
    'key,kwargs', [
        ('Ni_gr', {'mode': 'line', 'text': 'Ni', 'xy_kwargs': {'color': 'black'}}),
        ('Ni_gr', {'mode': 'line', 'legends': ['Ni'], 'xy_kwargs': {'color': 'black'}}),
        ('Ni_fgr', {'mode': 'fit', 'text': 'Ni', 'xy_kwargs': {'color': 'black'}})
    ]
)
def test_visualize(test_data, key, kwargs):
    fig = plt.figure()
    ax = fig.add_subplot(111)
    vis.visualize(test_data[key], ax=ax, **kwargs)
    plt.show(block=False)
    plt.close()



import typing as tp
from collections import namedtuple
from configparser import ConfigParser
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import scipy.optimize as opt
from bluesky.callbacks.stream import LiveDispatcher
from diffpy.pdfgetx import PDFGetter, PDFConfig
from event_model import RunRouter
from ophyd.sim import NumpySeqHandler

import pdfstream.units as units
from pdfstream.callbacks.basic import LiveWaterfall, NumpyExporter
from pdfstream.servers import CONFIG_DIR, ServerNames
from pdfstream.servers.base import ServerConfig, find_cfg_file, BaseServer


class LSQConfig(ConfigParser):
    """The configuration to make LSQ run router."""

    def __init__(self, *args, **kwargs):
        super(LSQConfig, self).__init__(*args, **kwargs)
        self._figs = list()

    @property
    def data_config(self):
        section = self["DATA KEYS"]
        tup = namedtuple(
            "data_config",
            ["x", "y", "lsq_type", "comp_id", "lsq_comps", "composition"]
        )
        return tup(
            section.get("x"),
            section.get("y"),
            section.get("lsq_type"),
            section.get("comp_id"),
            section.get("lsq_comps"),
            section.get("composition")
        )

    @property
    def interp_config(self):
        section = self["INTERPOLATION"]
        tup = namedtuple("interp_config", ["x"])
        tup.x = np.arange(
            section.getfloat("xmin"),
            section.getfloat("xmax"),
            section.getfloat("xstep")
        )
        return tup

    @property
    def fit_config(self):
        section = self["FITTING"]
        tup = namedtuple("fit_config", ["p"])
        tup.p = section.getfloat("p")
        return tup

    @property
    def trans_config(self):
        section = self["TRANSFORMATION"]
        return dict(
            dataformat=section.get("dataformat"),
            qmin=section.getfloat("qmin"),
            qmax=section.getfloat("qmax"),
            qmaxinst=section.getfloat("qmaxinst"),
            rpoly=section.getfloat("rpoly"),
            rmin=section.getfloat("rmin"),
            rmax=section.getfloat("rmax"),
            rstep=section.getfloat("rstep")
        )

    @property
    def export_config(self):
        section = self["EXPORTATION"]
        tup = namedtuple("export_config", ["directory", "file_prefix"])
        tup.directory = Path(section.get("directory"))
        tup.file_prefix = section.get("file_prefix")
        return tup

    @property
    def verbose(self):
        return self.getint("BASIC", "verbose", fallback=1)


class LSQServerConfig(LSQConfig, ServerConfig):
    """The configuration to make LSQ server."""
    pass


class LSQServer(BaseServer):
    """The server that decomposes one array data to a linear combination of the several array data in the record
    and save the data and visualize the residual of the decomposition and the PDF data transformed from it."""

    def __init__(self, config: LSQServerConfig):
        super(LSQServer, self).__init__(config.address, prefix=config.prefix)
        self.subscribe(LSQRunRouter(config))

    @classmethod
    def from_config(cls, config: LSQServerConfig):
        return LSQServer(config)

    @classmethod
    def from_cfg_file(cls, cfg_file: str):
        config = LSQServerConfig()
        config.read(cfg_file)
        return LSQServer(config)


class LSQRunRouter(RunRouter):
    """The run router that send the documents that can be processed by the decomposer."""

    def __init__(self, config: LSQConfig):
        factory = LSQFactory(config)
        super(LSQRunRouter, self).__init__([factory], handler_registry={"NPY_SEQ": NumpySeqHandler})


class LSQFactory:
    """The factory to generate the callbacks for the lsq run router."""

    def __init__(self, config: LSQConfig):
        self.config = config
        self.decomposer = LiveDecomposer(config)
        self.decomposer.subscribe(Visualiser(config))
        self.decomposer.subscribe(Exporter(config))

    def __call__(self, name, doc):
        if name == "start" and doc.get(self.config.data_config.lsq_type) is not None:
            return [self.decomposer], []
        return [], []


class LiveDecomposer(LiveDispatcher):
    """A callback to decompose one array data to a linear combination of the several array data in the record
    and transform the data to PDF."""

    def __init__(self, config: LSQConfig):
        super(LiveDecomposer, self).__init__()
        self.config = config
        self.memory = dict()
        self.cache = dict()

    def start(self, doc, _md=None):
        self.cache = dict()
        keys = self.config.data_config
        # get the type of the data, target, components or neither
        self.cache["lsq_type"] = doc.get(keys.lsq_type)
        # if it is target, there is a list of keys in memory, pointing to the data for decomposition
        self.cache["lsq_comps"] = doc.get(keys.lsq_comps, [])
        # if it is component, use the sample as the key in memory
        self.cache["comp_id"] = doc.get(keys.comp_id, "")
        # sample composition for pdfconfig
        self.cache["composition"] = doc.get(keys.composition, "C")
        super(LiveDecomposer, self).start(doc, _md=_md)

    def event(self, doc, _md=None):
        # if target, process the data and emit descriptor and events
        if self.cache["lsq_type"] == "target":
            if self.config.verbose > 0:
                print("Process data from the {}.".format(self.cache["comp_id"]))
            data = process_data(
                data=doc["data"],
                memory=self.memory,
                lsq_comps=self.cache["lsq_comps"],
                composition=self.cache["composition"],
                config=self.config
            )
            self.process_event({"data": data, 'descriptor': doc["descriptor"]})
        # if components, record the interpreted data in memory
        elif self.cache["lsq_type"] == "component":
            if self.config.verbose > 0:
                print("Record data from the {} in memory.".format(self.cache["comp_id"]))
            self.memory[self.cache["comp_id"]] = get_interp_data(doc["data"], self.config)
        # else do nothing
        else:
            pass

    def stop(self, doc, _md=None):
        self.cache = dict()
        super(LiveDecomposer, self).stop(doc, _md=_md)


def process_data(
    data: tp.Dict[str, np.ndarray],
    memory: tp.Dict[str, np.ndarray],
    lsq_comps: tp.List[str],
    composition: str,
    config: LSQConfig
) -> tp.Dict[str, np.ndarray]:
    """Process the data from the event."""
    # the interpolated target y data
    y = get_interp_data(data, config)
    # the component matrix where each row is a component data
    x = np.stack([memory[k] for k in lsq_comps], axis=0)
    # the model to decompose to a weight and the component matrix.
    model = Model(y, x, config.fit_config.p, xgrid=config.interp_config.x)
    opt.least_squares(model.cost_func, x0=model.x0)
    # transfer the data to PDF
    pdfconfig = PDFConfig(**config.trans_config, composition=composition)
    pdfgetter = PDFGetter(pdfconfig)
    pdfgetter(model.xgrid, model.yres)
    return {
        "y": model.y,
        "x": model.x,
        "w": model.w,
        "xgrid": model.xgrid,
        "yres": model.yres,
        "r": pdfgetter.gr[0],
        "g": pdfgetter.gr[1]
    }


def get_interp_data(data: tp.Dict[str, np.ndarray], config: LSQConfig) -> np.ndarray:
    """Get the x and y data from the data in event."""
    xp = data[config.data_config.x]
    yp = data[config.data_config.y]
    x = config.interp_config.x
    y = np.interp(x, xp, yp)
    return y


class Model:
    """The calculate Y - P(wX^T). Y is the a vector, P is a mapping from vector to vector, Pv = v'. v'_i =
    sqrt(p) v_i if v_i > 0 else v'_i = sqrt(1 - p) v_i. w is the vector and X is a matrix."""

    def __init__(self, y: np.ndarray, x: np.ndarray, p: float, xgrid=None):
        self.y = y
        self.x = x
        self.ycalc = None
        self.w = None
        self.yres = None
        self.xgrid = xgrid
        self.pos = np.sqrt(p)
        self.neg = np.sqrt(1 - p)

    def cost_func(self, w: np.ndarray):
        self.w = w
        self.ycalc = np.dot(w, self.x)
        self.yres = self.y - self.ycalc
        return np.where(self.yres > 0, self.pos * self.yres, self.neg * self.yres)

    @property
    def x0(self):
        return np.zeros(self.x.shape[0])


class Visualiser(RunRouter):
    """The callback to visualize documents from the decomposer."""

    def __init__(self, config: LSQConfig):
        factory = VisualiserFactory(config)
        super(Visualiser, self).__init__([factory], handler_registry={"NPY_SEQ": NumpySeqHandler})


class VisualiserFactory:
    """The factory of the visualizer."""

    def __init__(self, config: LSQConfig):
        self.config = config
        fig = plt.figure()
        self.callbacks = [
            LiveWaterfall(
                "xgrid",
                "yres",
                xlabel=units.LABELS.iq[0],
                ylabel=units.LABELS.iq[1],
                ax=fig.add_subplot(121)
            ),
            LiveWaterfall(
                "r",
                "g",
                xlabel=units.LABELS.gr[0],
                ylabel=units.LABELS.gr[1],
                ax=fig.add_subplot(122)
            )
        ]
        fig.show()

    def __call__(self, name, doc):
        if name == "start" and doc.get(self.config.data_config.lsq_type) == "target":
            return self.callbacks, []
        return [], []


class Exporter(RunRouter):
    """The callback to export documents from the decomposer."""

    def __init__(self, config: LSQConfig):
        factory = ExporterFactory(config)
        super(Exporter, self).__init__([factory], handler_registry={"NPY_SEQ": NumpySeqHandler})


class ExporterFactory:
    """The factory of the exporter."""

    def __init__(self, config: LSQConfig):
        self.config = config
        self.callbacks = [
            NumpyExporter(
                str(config.export_config.directory),
                file_prefix=config.export_config.file_prefix,
                data_keys=["y", "x", "w", "xgrid", "yres", "r", "g"]
            )
        ]

    def __call__(self, name, doc):
        if name == "start" and doc.get(self.config.data_config.lsq_type) == "target":
            return self.callbacks, []
        return [], []


def make_and_run(cfg_file: str = None):
    """Make and run LSQ server."""
    if not cfg_file:
        cfg_file = find_cfg_file(CONFIG_DIR, ServerNames.lsq)
    server = LSQServer.from_cfg_file(cfg_file)
    server.install_qt_kicker()
    server.start()
    return



import matplotlib.pyplot as plt
import numpy as np
import pytest

import pdfstream.integration.tools
import pdfstream.integration.tools as tools
from pdfstream.integration.tools import integrate


def test_bg_sub_error():
    with pytest.raises(ValueError):
        tools.bg_sub(np.ones((2, 2)), np.zeros((3, 3)))


@pytest.mark.parametrize(
    "case", [0]
)
def test_integrate(test_data, case):
    if case == 0:
        chi, setting = integrate(
            test_data["Ni_img"], test_data["ai"], mask=np.ones_like(test_data["Ni_img"]),
            integ_setting={"npt": 1000}
        )
        expect = np.zeros(1000)
        assert chi.shape == (2, 1000)
        assert np.array_equal(chi[1], expect)


@pytest.fixture
def user_mask(request, test_data):
    if request.param == "ones":
        return np.ones_like(test_data["Ni_img"])
    if request.param == "zeros":
        return np.zeros_like(test_data["Ni_img"])
    return None


@pytest.mark.parametrize(
    "user_mask, mask_setting",
    [
        (None, None),
        (None, {"alpha": 2, "upper_thresh": 1000}),
        ("ones", None),
        ("zeros", None)
    ],
    indirect=["user_mask"]
)
def test_auto_mask(test_data, user_mask, mask_setting):
    mask, _ = pdfstream.integration.tools.auto_mask(test_data['Ni_img'], test_data['ai'], user_mask=user_mask,
                                                    mask_setting=mask_setting)
    plt.matshow(mask)
    plt.colorbar()
    plt.show(block=False)
    plt.close()



import numpy
import pytest

import pdfstream.callbacks.from_event as mod

EVENT0 = {
    'time': [0.0],
    'uid': ['6a7a00e2-2a64-4284-9d67-e499ab682eea'],
    'seq_num': [1],
    'descriptor': '7711ee16-4b89-483b-9e96-94eb661cd4ee',
    'filled': {'pe1_image': ['d82442a7-6691-48bc-9c09-ae1c30e0aee4/0']},
    'data': {
        'pe1_image': [numpy.ones((3, 3))],
        'pe1_stats1_total': [9.0]
    }
}


@pytest.mark.parametrize(
    "event, det_name, expect",
    [
        (EVENT0, "pe1_image", numpy.ones((3, 3))),
        pytest.param(EVENT0, "pe1_stats1_total", None, marks=pytest.mark.xfail)
    ]
)
def test_get_image_from_event(event, det_name, expect):
    img = mod.get_image_from_event(event, det_name=det_name)
    assert numpy.array_equal(img, expect)



"""Get data from event."""
import typing

import numpy
from numpy import ndarray


def get_image_from_event(
    event: typing.Dict[str, typing.Any],
    det_name: str
) -> numpy.ndarray:
    """Read the image data from the event. It will be transferred to a numpy array.

    Parameters
    ----------
    event :
        The event document.

    det_name :
        The name of the key to the image. Usually, the detector name.

    Returns
    -------
    img :
        The two dimensional array of image.
    """
    data = event['data'][det_name]
    return squeeze_to_image(data)


def squeeze_to_image(
    arr: typing.Union[list, ndarray],
    ndim: int = 2
) -> ndarray:
    """Squeez the array to a 2d image array."""
    arr1 = numpy.asarray(arr)
    img = numpy.squeeze(arr1)
    if img.ndim != ndim:
        raise ValueError("Invalid image dimension. Expect {} but this is {}".format(ndim, img.ndim))
    return img



import numpy
import pytest

from pdfstream.transformation.tools import make_pdfgetter, use_pdfgetter


def test_make_pdfgetter(test_data):
    make_pdfgetter(
        test_data['Ni_config'], {'qmin': 1.}
    )


def test_use_pdfgetter(test_data):
    use_pdfgetter(test_data['Ni_chi'], test_data['Ni_pdfgetter'])


@pytest.mark.parametrize(
    "wrong_data",
    [
        numpy.zeros((2, 5, 1)),
        numpy.zeros((5, 2))
    ]
)
def test_use_pdfgetter_error(test_data, wrong_data):
    with pytest.raises(ValueError):
        use_pdfgetter(wrong_data, test_data['Ni_pdfgetter'])



import typing

from diffpy.pdfgetx import PDFConfig, PDFGetter
from numpy import ndarray

from pdfstream.transformation.tools import make_pdfgetter, use_pdfgetter, visualize

__all__ = [
    'get_pdf'
]


def get_pdf(pdfconfig: PDFConfig, chi: ndarray, plot_setting: typing.Union[str, dict] = None) -> PDFGetter:
    """Get the PDF from the chi data.

    The pdfgetter will be initiated by pdfconfig. It takes the chi data and does the calculation. The results
    will be visualized.

    Parameters
    ----------
    pdfconfig : PDFConfig
        This class stores all configuration data needed for generating PDF. See diffpy.pdfgetx.PDFConfig.

    chi : ndarray
        The array of chi data. The first row is Q or two theta, the second row is intensity.

    plot_setting : dict or 'OFF'
        The kwargs of plotting. See matplotlib.pyplot.plot. If 'OFF', skip visualization.

    test : bool
        If True, run in test mode.

    Returns
    -------
    pdfgetter :
        The PDFGetter.
    """
    pdfgetter = make_pdfgetter(pdfconfig)
    use_pdfgetter(chi, pdfgetter)
    if plot_setting != "OFF":
        visualize(pdfgetter, plot_setting)
    return pdfgetter



import pdfstream.callbacks.basic as mod


def test_NumpyExporter(tmpdir, array_stream):
    cb = mod.NumpyExporter(str(tmpdir), file_prefix="{start[sample_name]}_")
    for name, doc in array_stream:
        cb(name, doc)
    assert len(tmpdir.listdir()) > 0


def test_StackedNumpyExporter(tmpdir, array_stream):
    cb = mod.StackedNumpyExporter(str(tmpdir), file_prefix="{start[sample_name]}_", data_keys=["x0"])
    for name, doc in array_stream:
        cb(name, doc)
    assert len(tmpdir.listdir()) > 0


def test_StackedNumpyTextExporter(tmpdir, array_stream):
    cb = mod.StackedNumpyTextExporter(str(tmpdir), file_prefix="{start[sample_name]}_", data_keys=["x0"])
    for name, doc in array_stream:
        cb(name, doc)
    assert len(tmpdir.listdir()) > 0


def test_DataFrameExporter(tmpdir, array_stream):
    cb = mod.DataFrameExporter(str(tmpdir), file_prefix="{start[sample_name]}_")
    for name, doc in array_stream:
        cb(name, doc)
    assert len(tmpdir.listdir()) > 0


def test_SmartScalarPlot(ymax_stream):
    cb = mod.SmartScalarPlot("ymax")
    for name, doc in ymax_stream:
        cb(name, doc)
    cb.show()



from pkg_resources import resource_filename

import pdfstream.analyzers.xpd_analyzer as mod

fn = resource_filename("tests", "configs/xpd_server.ini")


def test_XPDAnalyzer(db_with_img_and_bg_img, tmpdir):
    raw_db = db_with_img_and_bg_img
    config = mod.XPDAnalyzerConfig()
    config.add_section("DATABASE")
    config.raw_db = raw_db
    config.add_section("FILE SYSTEM")
    config.tiff_base = str(tmpdir)
    config.calib_base = str(tmpdir)
    config.read(fn)
    analyzer = mod.XPDAnalyzer(config)
    run = raw_db[-1]
    analyzer.analyze(run)
    an_db = config.an_db
    mod.replay(an_db[-1])



import matplotlib.pyplot as plt
import numpy as np
import pytest

import pdfstream.integration.main as integ


@pytest.mark.parametrize(
    "img_key, mask, dk_img, bg_img, kwargs",
    [
        ('Ni_img', None, None, None, {}),
        ('Kapton_img', None, None, None, {}),
        (
            'Ni_img',
            None,
            None,
            'Kapton_img',
            {'bg_scale': 0.001, 'plot_setting': 'OFF', 'img_setting': 'OFF', 'mask_setting': 'OFF'}),
        (
            'Ni_img',
            None,
            None,
            None,
            {'mask_setting': {"alpha": 3}, 'plot_setting': 'OFF', 'img_setting': 'OFF'}
        ),
        (
            'Ni_img',
            'mask',
            None,
            None,
            {'mask_setting': "OFF", 'plot_setting': 'OFF', 'img_setting': 'OFF'}
        ),
        (
            'Ni_img',
            None,
            None,
            None,
            {
                'integ_setting': {'npt': 1024}, 'plot_setting': 'OFF', 'img_setting': 'OFF', 'mask_setting': 'OFF'
            }
        ),
        (
            'Ni_img',
            None,
            'Ni_img',
            None,
            {
                'integ_setting': {'npt': 1024}, 'plot_setting': 'OFF', 'img_setting': 'OFF', 'mask_setting': 'OFF'
            }
        )
    ]
)
def test_get_chi(test_data, img_key, mask, dk_img, bg_img, kwargs):
    integ.get_chi(
        test_data['ai'],
        test_data[img_key],
        dk_img=test_data.get(dk_img, None),
        bg_img=test_data.get(bg_img, None),
        **kwargs
    )
    plt.close()


@pytest.mark.parametrize(
    "img, bg_img, dk_img",
    [
        ("black_img", None, None),
        ("white_img", "white_img", None),
        ("white_img", None, "white_img"),
    ]
)
def test_get_chi_sanity(test_data, img, bg_img, dk_img):
    npt = 64
    result = integ.get_chi(
        test_data['ai'],
        test_data.get(img, None),
        dk_img=test_data.get(dk_img, None),
        bg_img=test_data.get(bg_img, None),
        mask_setting="OFF",
        integ_setting={'npt': npt},
        plot_setting="OFF",
        img_setting="OFF"
    )
    chi = result[0]
    assert np.array_equal(chi[1], np.zeros(npt))
    plt.close()


def test_avg_imgs(test_data):
    res = integ.avg_imgs([test_data['white_img'], test_data['white_img']], weights=[1, 1])
    assert np.array_equal(res, test_data['white_img'])



import pytest

import pdfstream.callbacks.composer as mod


@pytest.fixture
def gen_stream_args(test_data):
    data = {
        "img": test_data['Ni_img'],
        "motor": 0.0,
        "count": 1,
        "is_img": True,
        "name": "image"
    }
    metadata = {"hints": {"dimensions": [(["motor"], "primary")]}}
    return [data], metadata


def test_gen_stream(gen_stream_args):
    stream = mod.gen_stream(*gen_stream_args)
    for name, doc in stream:
        print(name, doc)



"""The functions used in the integration pipelines. All functions consume namespace and return the modified
namespace. """
from typing import Tuple

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.axes import Axes
from numpy import ndarray
from pyFAI.azimuthalIntegrator import AzimuthalIntegrator

from pdfstream.vend.masking import generate_binner, mask_img

INTEG_SETTING = dict(
    npt=1480,
    correctSolidAngle=False,
    method='splitpixel',
    unit='q_A^-1',
    safe=False
)
# default visualization setting
_LABEL = {
    "q_A^-1": r"Q ($\mathrm{\AA}^{-1}$)",
    "q_nm^-1": r"Q ({nm}$^{-1}$)",
    "2th_deg": r"2$\theta$ (deg)",
    "2th_rad": r"2$\theta$ (rad)",
    "r_mm": r"radius (mm)"
}


def bg_sub(img: ndarray, bg_img: ndarray, bg_scale: float = None) -> ndarray:
    """Subtract the background image from the data image inplace.

    Parameters
    ----------
    img : ndarray
        The 2D diffraction image array.

    bg_img : ndarray
        The 2D background image array.

    bg_scale : float
        The scale of the the background image.
    """
    if bg_scale is None:
        bg_scale = 1.
    if bg_img.shape != img.shape:
        raise ValueError(f"Unmatched shape between two images: {bg_img.shape}, {img.shape}.")
    return img - bg_scale * bg_img


def integrate(
    img: ndarray, ai: AzimuthalIntegrator, mask: ndarray = None, integ_setting: dict = None
) -> Tuple[ndarray, dict]:
    """Use AzimuthalIntegrator to integrate the image.

    Parameters
    ----------
    img : ndarray
        The 2D diffraction image array.

    ai : AzimuthalIntegrator
        The AzimuthalIntegrator instance.

    mask : ndarray
        The mask as a 0 and 1 array. 0 pixels are good pixels, 1 pixels are masked out.

    integ_setting : dict
        The user's modification to integration settings.

    Returns
    -------
    chi : ndarray
        The chi data. The first row is bin centers and the second row is the average intensity in bins.

    _integ_setting: dict
        The whole integration setting.
    """
    # merge integrate setting
    _integ_setting = INTEG_SETTING.copy()
    if integ_setting is not None:
        _integ_setting.update(integ_setting)
    # integrate
    xy = ai.integrate1d(img, mask=mask, **_integ_setting)
    chi = np.stack(xy)
    return chi, _integ_setting


def vis_img(img: ndarray, mask: ndarray = None, img_setting: dict = None, show: bool = True) -> Axes:
    """Visualize the processed image. The color map will be determined by statistics of the pixel values. The color map
    is determined by mean +/- z_score * std.

    Parameters
    ----------
    img : ndarray
        The 2D diffraction image array.

    mask: ndarray
        The whole integration setting.

    img_setting : dict
        The user's modification to imshow kwargs except a special key 'z_score'.

    show : bool
        If True, show the figure.

    Returns
    -------
    ax : Axes
        The axes with the image shown.
    """
    if img_setting is None:
        img_setting = dict()
    fig = plt.figure()
    ax: Axes = fig.add_subplot(111)
    if mask is not None:
        img = np.ma.masked_array(img, mask)
    mean, std = img.mean(), img.std()
    z_score = img_setting.pop('z_score', 2.)
    kwargs = {
        'vmin': mean - z_score * std,
        'vmax': mean + z_score * std
    }
    kwargs.update(**img_setting)
    img_obj = ax.matshow(img, **kwargs)
    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_xticklabels([])
    ax.set_yticklabels([])
    # color bar with magical settings to make it same size as the plot
    plt.colorbar(img_obj, ax=ax, fraction=0.046, pad=0.04)
    if show:
        plt.show(block=False)
    return ax


def vis_chi(chi: ndarray, plot_setting: dict = None, unit: str = None, show: bool = True) -> Axes:
    """Visualize the chi curve.

    Parameters
    ----------
    chi : ndarray
        The chi data. The first row is bin centers and the second row is the average intensity in bins.

    plot_setting : dict
        The kwargs for the plot function.

    unit: str
        The unit of the chi data. It affects the label of the plot. If None, no unit.

    show : bool
        If True, show the figure.

    Returns
    -------
    ax : Axes
        The axes with the curve plotted.
    """
    if plot_setting is None:
        plot_setting = dict()
    fig = plt.figure()
    ax = fig.add_subplot(111)
    ax.plot(chi[0], chi[1], **plot_setting)
    if unit:
        ax.set_xlabel(_LABEL.get(unit))
    ax.set_ylabel('I (A. U.)')
    if show:
        plt.show(block=False)
    return ax


def auto_mask(
    img: ndarray,
    ai: AzimuthalIntegrator,
    user_mask: ndarray = None,
    mask_setting: dict = None
) -> Tuple[ndarray, dict]:
    """Automatically generate the mask of the image.

    Parameters
    ----------
    img : ndarray
        The 2D diffraction image array.

    ai : AzimuthalIntegrator
        The AzimuthalIntegrator instance.

    mask_setting : dict
        The user's modification to auto-masking settings.

    user_mask : ndarray
        A mask provided by user. It is an integer array. 0 are good pixels, 1 are masked out.

    Returns
    -------
    mask : ndarray
        The mask as an integer array. 0 are good pixels, 1 are masked out.

    _mask_setting : dict
        The whole mask_setting.
    """
    if mask_setting is not None:
        _mask_setting = mask_setting
    else:
        _mask_setting = dict()
    binner = generate_binner(ai, img.shape)
    tmsk = np.invert(user_mask.astype(bool)) if user_mask is not None else None
    mask = mask_img(img, binner, tmsk=tmsk, **_mask_setting)
    mask = np.invert(mask).astype(int)
    return mask, _mask_setting



import pytest

import pdfstream.utils.dct_ops as dct_ops


def test_paths():
    dct = {
        "k1": {
            "k11": "v11",
            "k12": {
                "k121": "v121"
            }
        },
        "k2": ["v21", "v22"]
    }
    assert list(dct_ops.paths(dct)) == [
        ("k1", "k11", "v11"),
        ("k1", "k12", "k121", "v121"),
        ("k2", 0, "v21"),
        ("k2", 1, "v22")
    ]


def test_to_dict(test_data):
    dct_ops.to_dict(test_data["Ni_config"])


@pytest.mark.parametrize(
    "dct, op, expect",
    [
        (
            {"k0": {"k1": 0, "k2": 1}},
            lambda x: x + 1,
            {"k0": {"k1": 1, "k2": 2}}
        )
    ]
)
def test_iter_dct(dct, op, expect):
    dct1 = dct_ops.iter_dct(dct, op)
    assert dct1 == expect


@pytest.mark.parametrize(
    "dct, keys, expect",
    [
        ({"k": {"k1": "v"}}, ("k", "k1"), "v"),
        ({"k": "v"}, tuple(), {"k": "v"}),
    ]
)
def test_get_value(dct, keys, expect):
    value = dct_ops.get_value(dct, keys=keys)
    assert value == expect



import typing as tp
from pathlib import Path

import numpy as np
import pandas as pd
from bluesky.callbacks import CallbackBase
from bluesky.callbacks.best_effort import LivePlot, LiveScatter
from bluesky.callbacks.broker import LiveImage
from databroker.v2 import Broker
from event_model import unpack_event_page
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.widgets import Slider
from xpdview.waterfall import Waterfall

import pdfstream.callbacks.from_descriptor as fd
import pdfstream.callbacks.from_start as fs
from pdfstream.vend.formatters import SpecialStr


class ArrayExporter(CallbackBase):
    """An base class for the callbacks to find and export the 1d array."""
    file_suffix = ""
    file_stem = ""

    def __init__(self, directory: str, *, file_prefix: str, data_keys: tp.List[str] = None):
        super(ArrayExporter, self).__init__()
        self.directory = Path(directory)
        self.directory.mkdir(parents=True, exist_ok=True)
        self.file_template = SpecialStr(
            file_prefix + self.file_stem + self.file_suffix
        )
        self.data_keys = data_keys
        self.start_doc = None
        self.descriptor_doc = None

    def start(self, doc):
        self.start_doc = doc
        super(ArrayExporter, self).start(doc)

    def descriptor(self, doc):
        if not self.data_keys:
            self.data_keys = list(fd.yield_1d_array(doc["data_keys"]))
        self.descriptor_doc = doc
        super(ArrayExporter, self).descriptor(doc)

    def event(self, doc):
        self.export(doc)
        super(ArrayExporter, self).event(doc)

    def event_page(self, doc):
        for event in unpack_event_page(doc):
            self.event(event)

    def stop(self, doc):
        self.start_doc = None
        self.descriptor_doc = None
        super(ArrayExporter, self).stop(doc)

    def export(self, doc):
        pass


class NumpyExporter(ArrayExporter):
    """An exporter to export the array data one by one in .npy file."""
    file_suffix = ".npy"
    file_stem = "{descriptor[name]}-{field}-{event[seq_num]}"

    def export(self, doc):
        for data_key in self.data_keys:
            arr: np.ndarray = doc["data"][data_key]
            filename = self.file_template.format(start=self.start_doc, descriptor=self.descriptor_doc, event=doc,
                                                 field=data_key)
            filepath = self.directory.joinpath(filename)
            np.save(str(filepath), arr)


class StackedNumpyExporter(ArrayExporter):
    """An exporter to export the column-stacked array data in .npy file."""
    file_suffix = ".npy"
    file_stem = "{descriptor[name]}-{field}-{event[seq_num]}"

    def export(self, doc):
        arr: np.ndarray = np.stack([doc["data"][data_key] for data_key in self.data_keys], axis=-1)
        field = "-".join(self.data_keys)
        filename = self.file_template.format(start=self.start_doc, descriptor=self.descriptor_doc, event=doc,
                                             field=field)
        filepath = self.directory.joinpath(filename)
        np.save(str(filepath), arr)


class StackedNumpyTextExporter(ArrayExporter):
    """An exporter to export the column-stacked array data in .txt file."""
    file_suffix = ".txt"
    file_stem = "{descriptor[name]}-{field}-{event[seq_num]}"

    def export(self, doc):
        arr: np.ndarray = np.stack([doc["data"][data_key] for data_key in self.data_keys], axis=-1)
        field = "-".join(self.data_keys)
        filename = self.file_template.format(start=self.start_doc, descriptor=self.descriptor_doc, event=doc,
                                             field=field)
        filepath = self.directory.joinpath(filename)
        header = " ".join(self.data_keys)
        np.savetxt(str(filepath), arr, header=header)


class DataFrameExporter(ArrayExporter):
    """An exporter to export data in a dataframe in the .csv file."""
    file_suffix = ".csv"
    file_stem = "{descriptor[name]}-{field}-{event[seq_num]}"

    def export(self, doc):
        _data = {data_key: pd.Series(doc["data"][data_key]) for data_key in self.data_keys}
        df = pd.DataFrame(data=_data)
        filename = self.file_template.format(start=self.start_doc, descriptor=self.descriptor_doc, event=doc,
                                             field="data")
        filepath = self.directory.joinpath(filename)
        df.to_csv(str(filepath))


class MyLiveImage(LiveImage):
    """A customized LiveImage."""

    def show(self):
        self.cs._fig.show()


class LiveMaskedImage(MyLiveImage):
    """Live image show of a image with a mask."""

    def __init__(self, field: str, msk_field: str, *, cmap: str, norm: tp.Callable = None,
                 limit_func: tp.Callable = None, auto_draw: bool = True, interpolation: str = None,
                 window_title: str = None, db: Broker = None):
        self.msk_field = msk_field
        super(LiveMaskedImage, self).__init__(
            field, cmap=cmap, norm=norm, limit_func=limit_func,
            auto_redraw=auto_draw, interpolation=interpolation, window_title=window_title, db=db
        )

    def event(self, doc):
        super(LiveImage, self).event(doc)
        data = np.ma.masked_array(doc["data"][self.field], doc["data"][self.msk_field])
        self.update(data)


class MyWaterfall(Waterfall):
    """An adaptation of WaterFall. Allow using ax instead of Figure."""

    def __init__(self, *, xlabel: str, ylabel: str, ax: Axes, **kwargs):
        super(Waterfall, self).__init__()
        self.ax = ax
        self.fig = self.ax.figure
        self.canvas = self.fig.canvas
        self.kwargs = kwargs
        self.x_array_list = []
        self.y_array_list = []

        # callback for showing legend
        self.canvas.mpl_connect("pick_event", self.on_plot_hover)
        self.key_list = []
        self.unit = (xlabel, ylabel)

        # add sliders, which store information
        self.ydist = 0
        self.xdist = 0

        y_offset_slider_ax = self.fig.add_axes([0.15, 0.95, 0.25, 0.035])
        self.y_offset_slider = Slider(
            y_offset_slider_ax,
            "y-offset",
            0.0,
            1.0,
            valinit=0.1,
            valfmt="%1.2f",
        )
        self.y_offset_slider.on_changed(self.update_y_offset)

        x_offset_slider_ax = self.fig.add_axes([0.6, 0.95, 0.25, 0.035])
        self.x_offset_slider = Slider(
            x_offset_slider_ax,
            "x-offset",
            0.0,
            1.0,
            valinit=0.,
            valfmt="%1.2f",
        )
        self.x_offset_slider.on_changed(self.update_x_offset)


class LiveWaterfall(CallbackBase):
    """A live water plot for the one dimensional data."""

    def __init__(self, x: str, y: str, *, xlabel: str, ylabel: str, ax: Axes, **kwargs):
        """Initiate the instance.

        Parameters
        ----------
        x :
            The key of the independent variable.

        y :
            The key of the dependent variable.

        xlabel :
            The tuple of the labels of x shown in the figure.

        ylabel :
            The tuple of the labels of y shown in the figure.

        ax :
            The axes to plot.

        kwargs :
            The kwargs for the matplotlib.pyplot.plot.
        """
        super().__init__()
        self.x = x
        self.y = y
        self.ax = ax
        self.waterfall = MyWaterfall(xlabel=xlabel, ylabel=ylabel, ax=self.ax, **kwargs)

    def start(self, doc):
        super(LiveWaterfall, self).start(doc)
        self.waterfall.clear()

    def event(self, doc):
        super(LiveWaterfall, self).event(doc)
        x_data = doc["data"][self.x]
        y_data = doc["data"][self.y]
        key = doc['seq_num']
        self.update(key, (x_data, y_data))

    def update(self, key: str, int_data: tp.Tuple[np.ndarray, np.ndarray]):
        self.waterfall.update(key_list=[key], int_data_list=[int_data])

    def show(self):
        self.ax.figure.show()


class SmartScalarPlot(CallbackBase):
    """A plot for scalar variable. Use LivePlot for one dimensional case and Use LiveScatter for two dimensional
    case """

    def __init__(self, y: str, *, ax: Axes = None, ylabel: str = None, **kwargs):
        super(SmartScalarPlot, self).__init__()
        self.y = y
        if ax is None:
            fig = plt.figure()
            ax = fig.add_subplot(111)
        self.ax = ax
        self.ylabel = ylabel
        self.kwargs = kwargs
        self.callback = None

    def start(self, doc):
        super(SmartScalarPlot, self).start(doc)
        self.clear()
        indeps = fs.get_indeps(doc, exclude={"time"})
        if len(indeps) == 1:
            self.callback = LivePlot(self.y, x=indeps.pop(), ax=self.ax, **self.kwargs)
        elif len(indeps) == 2:
            self.callback = LiveScatter(indeps.pop(), indeps.pop(), self.y, ax=self.ax, **self.kwargs)
        else:
            self.callback = LivePlot(self.y, ax=self.ax, **self.kwargs)
        self.callback.start(doc)

    def descriptor(self, doc):
        super(SmartScalarPlot, self).descriptor(doc)
        self.callback.descriptor(doc)

    def event(self, doc):
        super(SmartScalarPlot, self).event(doc)
        self.callback.event(doc)
        if self.ylabel:
            self.ax.set_ylabel(self.ylabel)

    def stop(self, doc):
        super(SmartScalarPlot, self).stop(doc)
        self.callback.stop(doc)

    def clear(self):
        self.ax.cla()
        self.callback = None

    def show(self):
        self.ax.figure.show()



import numpy as np
import pytest
from pkg_resources import resource_filename

import pdfstream.servers.lsq_server as mod
from pdfstream.callbacks.composer import gen_stream

cfg_file = resource_filename("tests", "configs/lsq_server.ini")


@pytest.fixture
def runs():
    x = np.arange(0, 26, 0.1)
    comp0 = gen_stream(
        [{"q": x, "iq": np.sin(x)}],
        {"lsq_type": "component", "sample_name": "c0"}
    )
    comp1 = gen_stream(
        [{"q": x, "iq": np.cos(x)}],
        {"lsq_type": "component", "sample_name": "c1"}
    )
    target = gen_stream(
        [{"q": x, "iq": np.cos(x) + np.sin(x) + np.exp(x / 26)}],
        {"lsq_type": "target", "lsq_comps": ["c0", "c1"], "sample_name": "c2"}
    )
    return comp0, comp1, target


def test_LSQRunRouter(runs, tmpdir):
    config = mod.LSQConfig()
    config.read(cfg_file)
    config.set("EXPORTATION", "directory", str(tmpdir))
    lsq_cb = mod.LSQRunRouter(config)
    for run in runs:
        for name, doc in run:
            lsq_cb(name, doc)
    assert len(tmpdir.listdir()) > 0


def test_LSQServer(tmpdir):
    config = mod.LSQServerConfig()
    config.read(cfg_file)
    config.set("EXPORTATION", "directory", str(tmpdir))
    mod.LSQServer.from_config(config)



from concurrent.futures import ThreadPoolExecutor, as_completed

import numpy as np
from skbeam.core.accumulators.binned_statistic import BinnedStatistic1D
from skbeam.core.mask import margin

from pdfstream.vend.jittools import mask_ring_mean, mask_ring_median

mask_ring_dict = {"median": mask_ring_median, "mean": mask_ring_mean}


def map_to_binner(pixel_map, bins, mask=None):
    """Transforms pixel map and bins into a binner

    Parameters
    ----------
    pixel_map: np.ndarray
        The map between pixels and values
    bins: np.ndarray
        The bins to use in the binner
    mask: np.ndarray, optional
        The mask for the pixel map

    Returns
    -------
    BinnedStatistic1D:
        The binner

    """
    if mask is not None:
        mask = mask.flatten()
    return BinnedStatistic1D(pixel_map.flatten(), bins=bins, mask=mask)


def generate_binner(geo, img_shape, mask=None):
    """Create a pixel resolution BinnedStats1D instance

    Parameters
    ----------
    geo : pyFAI.geometry.Geometry instance
        The calibrated geometry
    img_shape : tuple, optional
        The shape of the image, if None pull from the mask. Defaults to None.
    mask : np.ndarray, optional
        The mask to be applied, if None no mask is applied. Defaults to None.
    Returns
    -------
    BinnedStatistic1D :
        The configured instance of the binner.
    """
    return map_to_binner(*generate_map_bin(geo, img_shape), mask=mask)


def mask_img(
    img,
    binner,
    edge=30,
    lower_thresh=0.0,
    upper_thresh=None,
    alpha=2,
    auto_type="median",
    tmsk=None,
    pool=None,
):
    """
    Mask an image based off of various methods

    Parameters
    ----------
    img: np.ndarray
        The image to be masked
    binner : BinnedStatistic1D instance
        The binned statistics information
    edge: int, optional
        The number of edge pixels to mask. Defaults to 30. If None, no edge
        mask is applied
    lower_thresh: float, optional
        Pixels with values less than or equal to this threshold will be masked.
        Defaults to 0.0. If None, no lower threshold mask is applied
    upper_thresh: float, optional
        Pixels with values greater than or equal to this threshold will be
        masked.
        Defaults to None. If None, no upper threshold mask is applied.
    alpha: float, optional
        Then number of acceptable standard deviations, if tuple then we use
        a linear distribution of alphas from alpha[0] to alpha[1], if array
        then we just use that as the distribution of alphas. Defaults to 3.
        If None, no outlier masking applied.
    auto_type: {'median', 'mean'}, optional
        The type of binned outlier masking to be done, 'median' is faster,
        where 'mean' is more accurate, defaults to 'median'.
    tmsk: np.ndarray, optional
        The starting mask to be compounded on. Defaults to None. If None mask
        generated from scratch.
    pool : Executor instance
        A pool against which jobs can be submitted for parallel processing

    Returns
    -------
    tmsk: np.ndarray
        The mask as a boolean array. True pixels are good pixels, False pixels
        are masked out.

    """

    if tmsk is None:
        working_mask = np.ones(np.shape(img)).astype(bool)
    else:
        working_mask = tmsk.copy()
    if edge:
        working_mask *= margin(np.shape(img), edge)
    if lower_thresh is not None:
        working_mask *= (img >= lower_thresh).astype(bool)
    if upper_thresh is not None:
        working_mask *= (img <= upper_thresh).astype(bool)
    if alpha:
        working_mask *= binned_outlier(
            img,
            binner,
            alpha=alpha,
            tmsk=working_mask,
            mask_method=auto_type,
            pool=pool,
        )
    working_mask = working_mask.astype(np.bool)
    return working_mask


def binned_outlier(
    img, binner, tmsk, alpha=3, mask_method="median", pool=None
):
    """Sigma Clipping based masking.

    Parameters
    ----------
    img : np.ndarray
        The image
    binner : BinnedStatistic1D instance
        The binned statistics information
    alpha : float, optional
        The number of standard deviations to clip, defaults to 3
    tmsk : np.ndarray, optional
        Prior mask. If None don't use a prior mask, defaults to None.
    mask_method : {'median', 'mean'}, optional
        The method to use for creating the mask, median is faster, mean is more
        accurate. Defaults to median.
    pool : Executor instance
        A pool against which jobs can be submitted for parallel processing

    Returns
    -------
    np.ndarray:
        The mask
    """
    if pool is None:
        pool = ThreadPoolExecutor(max_workers=20)
    # skbeam 0.0.12 doesn't have argsort_index cached
    idx = binner.argsort_index
    tmsk = tmsk.flatten()
    tmsk2 = tmsk[idx]
    vfs = img.flatten()[idx]
    pfs = np.arange(np.size(img))[idx]
    t = []
    i = 0
    for k in binner.flatcount:
        m = tmsk2[i: i + k]
        vm = vfs[i: i + k][m]
        if k > 0 and len(vm) > 0:
            t.append((vm, (pfs[i: i + k][m]), alpha))
        i += k
    p_err = np.seterr(all="ignore")
    # only run tqdm on mean since it is slow
    with pool as p:
        futures = [
            p.submit(mask_ring_dict[mask_method], *x)
            for x in t
        ]
    removals = []
    for f in as_completed(futures):
        removals.extend(f.result())
    np.seterr(**p_err)
    tmsk[removals] = False
    tmsk = tmsk.reshape(np.shape(img))
    return tmsk.astype(bool)


def generate_map_bin(geo, img_shape):
    """Create a q map and the pixel resolution bins

    Parameters
    ----------
    geo : pyFAI.geometry.Geometry instance
        The calibrated geometry
    img_shape : tuple, optional
        The shape of the image, if None pull from the mask. Defaults to None.

    Returns
    -------
    q : ndarray
        The q map
    qbin : ndarray
        The pixel resolution bins
    """
    r = geo.rArray(img_shape)
    q = geo.qArray(img_shape) / 10  # type: np.ndarray
    q_dq = geo.deltaQ(img_shape) / 10  # type: np.ndarray

    pixel_size = [getattr(geo, a) for a in ["pixel1", "pixel2"]]
    rres = np.hypot(*pixel_size)
    rbins = np.arange(np.min(r) - rres / 2., np.max(r) + rres / 2., rres / 2.)
    rbinned = BinnedStatistic1D(r.ravel(), statistic=np.max, bins=rbins)

    qbin_sizes = rbinned(q_dq.ravel())
    qbin_sizes = np.nan_to_num(qbin_sizes)
    qbin = np.cumsum(qbin_sizes)
    qbin[0] = np.min(q_dq)
    if np.max(q) > qbin[-1]:
        qbin[-1] = np.max(q)
    return q, qbin



from pathlib import Path

import pytest
from pkg_resources import resource_filename

import pdfstream.servers
import pdfstream.servers.base as mod

fn = Path(resource_filename("tests", "configs"))


@pytest.mark.parametrize(
    "name", ["xpd", "lsq", pytest.param("unknown", marks=pytest.mark.xfail(raises=FileNotFoundError))]
)
def test_find_cfg_file(name):
    mod.find_cfg_file(fn, name)


def test_StartStopCallback(simple_stream):
    cb = pdfstream.servers.base.StartStopCallback()
    for name, doc in simple_stream:
        cb(name, doc)



"""Get data from the start and database."""
import itertools
import typing

import numpy
from databroker.core import BlueskyRunFromGenerator
from databroker.v2 import Broker
from numpy import ndarray

import pdfstream.io as io
from pdfstream.data import ni_dspacing_file


def query_ai(
    start: typing.Dict[str, typing.Any],
    calibration_md_key: str,
) -> typing.Union[None, typing.Dict[str, typing.Any]]:
    """Query the azimuthal integrator from the start document.

    If the poni_file is provided, use the poni file instead and ignore the information in start document.

    Parameters
    ----------
    start :
        The start document.

    calibration_md_key :
        The key chain to find the calibration metadata.

    Returns
    -------
    ai :
        The azimuthal integrator.
    """
    if calibration_md_key not in start:
        raise ValueError("Missing key {} in start document.".format(calibration_md_key))
    return io.load_ai_from_calib_result(start[calibration_md_key])


def query_bg_img(
    start: typing.Dict[str, typing.Any],
    bkgd_sample_name_key: str,
    sample_name_key: str,
    det_name: str,
    db: Broker = None,
    dk_id_key: str = None
) -> typing.Union[ndarray, None]:
    """Find background image from database according to the start document of a run.

    The function will find the background run id in the start and query this run in database. Then, it read
    the image data from the run. If dark image id is also provided, it will subtract the dark image from
    the background image. It assumes that the image is a single shot. If no background image id info in start
    document, return None.

    Parameters
    ----------
    start :
        The start document of the measurement run. It may contain the id of the background run.

    bkgd_sample_name_key :
        The key of the sample of background.

    sample_name_key :
        The key of the sample name.

    det_name :
        The name in the background image data in the xarray of the run.

    db :
        The database that contains the background image run.

    dk_id_key :
        The key of dark image id in the start document of background image run.

    Returns
    -------
    bg_img :
        The background image either dark subtracted or raw. If not found, None.
    """
    if db is None:
        return None
    sample_name = start.get(bkgd_sample_name_key)
    result = list(db.search({sample_name_key: sample_name}))
    if len(result) == 0:
        return None
    bg_run = db[result[-1]]
    img = get_img_from_run(bg_run, det_name=det_name)
    bg_start = get_start_of_run(bg_run)
    dk_img = query_dk_img(bg_start, det_name=det_name, db=db, dk_id_key=dk_id_key) if dk_id_key else None
    return numpy.subtract(img, dk_img) if dk_img is not None else img


def query_dk_img(
    start: typing.Dict[str, typing.Any],
    det_name: str,
    db: Broker = None,
    dk_id_key: str = None
) -> typing.Union[ndarray, None]:
    """Find the dark image according to the start document of a run.

    The function will find the dark id in the start document and search the dark run in database and read the
    image from the run. It assumes that the image is a single shot. If no dark id info in the start document,
    return None.

    Parameters
    ----------
    start :
        The start document of a run.

    det_name :
        The name in the background image data in the xarray of the run.

    db :
        The database that contains the background image run.

    dk_id_key :
        The key of dark image id in the start document of background image run.

    Returns
    -------
    dk_img :
        The raw dark image. If not found, None.
    """
    dk_run = get_dk_run(start, db, dk_id_key)
    if dk_run is None:
        return None
    return get_img_from_run(dk_run, det_name)


def get_dk_run(start: dict, db: Broker, dk_id_key: str) -> typing.Union[BlueskyRunFromGenerator, None]:
    """Get the dark image run id. If not found, return None."""
    if db and dk_id_key and dk_id_key in start:
        dk_id = start[dk_id_key]
        return db[dk_id]
    return None


def get_img_from_run(run: BlueskyRunFromGenerator, det_name: str) -> ndarray:
    """Read a single image of a detector from a run (databroker v2)."""
    ds = run.primary.read()
    img: ndarray = ds[det_name].values
    # remove all single dimensions
    img = numpy.squeeze(img)
    if img.ndim != 2:
        raise ValueError("Invalid number of dimension for an image: {}. Expect 2.".format(img.ndim))
    return img


def get_start_of_run(run: BlueskyRunFromGenerator):
    """Read the start document of a run (databroker v2)."""
    return run.metadata['start']


def query_bt_info(
    start: typing.Dict[str, typing.Any],
    composition_key: str,
    wavelength_key: str,
    default_composition: str = "Ni"
) -> dict:
    """Query the necessary information for the PDFGetter."""
    if composition_key in start:
        composition = start[composition_key]
        if isinstance(composition, dict):
            composition_str = "".join(["{}{}".format(k, v) for k, v in composition.items()])
        elif isinstance(composition, str):
            composition_str = composition
        else:
            raise ValueError("Cannot parse composition: {}".format(type(composition)))
    else:
        composition_str = default_composition
    if wavelength_key in start:
        wavelength = float(start[wavelength_key])
    else:
        wavelength = None
    return {
        "composition": composition_str,
        "wavelength": wavelength
    }


def get_indeps(start: dict, exclude: set = frozenset()) -> set:
    """Get independent variables from the hints in start."""
    return set(
        itertools.chain.from_iterable(
            [n for n, s in start.get("hints", {}).get("dimensions", [])]
        )
    ) - exclude


def get_calib_info(
    start: typing.Dict[str, typing.Any],
    detector_key: str,
    wavelength_key: str,
    calibrant_key: str
) -> typing.Dict[str, str]:
    """Get the information for pyfail calib2 gui in the start. If no such key, return empty string."""
    calibrant = str(start.get(calibrant_key, ""))
    # a special case for xpdacq
    calibrant = str(ni_dspacing_file) if calibrant in ("Ni_calib", "Ni") else calibrant
    return {
        "detector": str(start.get(detector_key, "")),
        "wavelength": str(start.get(wavelength_key, "")),
        "calibrant": calibrant
    }



import json
from pathlib import Path

from event_model import DocumentNames, _Validator
from pkg_resources import resource_filename as rs_fn

__all__ = [
    "analysis_in_schemas",
    "analysis_out_schemas",
    "Validator",
    "DocumentNames"
]

ANALYSIS_IN_SCHEMA_NAMES = {
    DocumentNames.start: 'schemas/analysis_in/run_start.json',
    DocumentNames.stop: 'schemas/analysis_in/run_stop.json',
    DocumentNames.event: 'schemas/analysis_in/event.json',
    DocumentNames.event_page: 'schemas/analysis_in/event_page.json',
    DocumentNames.descriptor: 'schemas/analysis_in/event_descriptor.json',
    DocumentNames.datum: 'schemas/analysis_in/datum.json',
    DocumentNames.datum_page: 'schemas/analysis_in/datum_page.json',
    DocumentNames.resource: 'schemas/analysis_in/resource.json'
}

analysis_in_schemas = {}
for doc_name, filename in ANALYSIS_IN_SCHEMA_NAMES.items():
    with Path(rs_fn('pdfstream', filename)).open("r") as fin:
        analysis_in_schemas[doc_name] = json.load(fin)

ANALYSIS_OUT_SCHEMA_NAMES = {
    DocumentNames.start: 'schemas/analysis_out/run_start.json',
    DocumentNames.stop: 'schemas/analysis_out/run_stop.json',
    DocumentNames.event: 'schemas/analysis_out/event.json',
    DocumentNames.event_page: 'schemas/analysis_out/event_page.json',
    DocumentNames.descriptor: 'schemas/analysis_out/event_descriptor.json',
    DocumentNames.datum: 'schemas/analysis_out/datum.json',
    DocumentNames.datum_page: 'schemas/analysis_out/datum_page.json',
    DocumentNames.resource: 'schemas/analysis_out/resource.json'
}

analysis_out_schemas = {}
for doc_name, filename in ANALYSIS_OUT_SCHEMA_NAMES.items():
    with Path(rs_fn('pdfstream', filename)).open("r") as fin:
        analysis_out_schemas[doc_name] = json.load(fin)


def print_data_keys(schemas: dict):
    """Print out data keys in event descriptor."""
    data_keys: dict = schemas[DocumentNames.descriptor]['properties']['data_keys']
    for k, v in data_keys.get("properties", {}).items():
        print("{}: {}".format(k, v.get("description", "")))


class Validator(object):
    """The json schema validator wrapped in a bluesky callback.

    Attributes
    ----------
    schemas : dict
        A mapping from the doc name to the schemas for that doc.
    """
    document_names = {
        "start": DocumentNames.start,
        "stop": DocumentNames.stop,
        "event": DocumentNames.event,
        "event_page": DocumentNames.event_page,
        "descriptor": DocumentNames.descriptor,
        "datum": DocumentNames.datum,
        "datum_page": DocumentNames.datum_page,
        "resource": DocumentNames.resource
    }

    def __init__(self, schemas: dict):
        self.schemas = schemas

    def __call__(self, name: str, doc: dict):
        doc_type = self.document_names[name]
        schema = self.schemas[doc_type]
        validator = _Validator(schema=schema)
        validator.validate(doc)
